import asyncio
import sys
import time
from unittest import mock

import pytest

from mcstatus.address import Address
from mcstatus.pinger import AsyncServerPinger
from mcstatus.protocol.connection import Connection


def async_decorator(f):
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


class FakeAsyncConnection(Connection):
    async def read_buffer(self):  # pyright: ignore[reportIncompatibleMethodOverride]
        return super().read_buffer()


class TestAsyncServerPinger:
    def setup_method(self):
        self.pinger = AsyncServerPinger(
            FakeAsyncConnection(),  # type: ignore[arg-type]
            address=Address("localhost", 25565),
            version=44,
        )

    def test_handshake(self):
        self.pinger.handshake()

        assert self.pinger.connection.flush() == bytearray.fromhex("0F002C096C6F63616C686F737463DD01")

    def test_read_status(self):
        self.pinger.connection.receive(
            bytearray.fromhex(
                "7200707B226465736372697074696F6E223A2241204D696E65637261667420536572766572222C22706C6179657273223A7B2"
                "26D6178223A32302C226F6E6C696E65223A307D2C2276657273696F6E223A7B226E616D65223A22312E382D70726531222C22"
                "70726F746F636F6C223A34347D7D"
            )
        )
        status = async_decorator(self.pinger.read_status)()

        assert status.raw == {
            "description": "A Minecraft Server",
            "players": {"max": 20, "online": 0},
            "version": {"name": "1.8-pre1", "protocol": 44},
        }
        assert self.pinger.connection.flush() == bytearray.fromhex("0100")

    def test_read_status_invalid_json(self):
        self.pinger.connection.receive(bytearray.fromhex("0300017B"))
        with pytest.raises(IOError):
            async_decorator(self.pinger.read_status)()

    def test_read_status_invalid_reply(self):
        self.pinger.connection.receive(
            bytearray.fromhex(
                "4F004D7B22706C6179657273223A7B226D6178223A32302C226F6E6C696E65223A307D2C2276657273696F6E223A7B226E616"
                "D65223A22312E382D70726531222C2270726F746F636F6C223A34347D7D"
            )
        )

        async_decorator(self.pinger.read_status)()

    def test_read_status_invalid_status(self):
        self.pinger.connection.receive(bytearray.fromhex("0105"))

        with pytest.raises(IOError):
            async_decorator(self.pinger.read_status)()

    def test_test_ping(self):
        self.pinger.connection.receive(bytearray.fromhex("09010000000000DD7D1C"))
        self.pinger.ping_token = 14515484

        assert async_decorator(self.pinger.test_ping)() >= 0
        assert self.pinger.connection.flush() == bytearray.fromhex("09010000000000DD7D1C")

    def test_test_ping_invalid(self):
        self.pinger.connection.receive(bytearray.fromhex("011F"))
        self.pinger.ping_token = 14515484

        with pytest.raises(IOError):
            async_decorator(self.pinger.test_ping)()

    def test_test_ping_wrong_token(self):
        self.pinger.connection.receive(bytearray.fromhex("09010000000000DD7D1C"))
        self.pinger.ping_token = 12345

        with pytest.raises(IOError):
            async_decorator(self.pinger.test_ping)()

    @pytest.mark.asyncio
    @pytest.mark.flaky(reruns=5, condition=sys.platform.startswith("win32"))
    async def test_latency_is_real_number(self):
        """``time.perf_counter`` returns fractional seconds, we must convert it to milliseconds."""

        def mocked_read_buffer():
            time.sleep(0.001)
            return mock.DEFAULT

        with mock.patch.object(FakeAsyncConnection, "read_buffer") as mocked:
            mocked.side_effect = mocked_read_buffer
            mocked.return_value.read_varint = lambda: 0  # overwrite `async` here
            mocked.return_value.read_utf = (
                lambda: """
            {
                "description": "A Minecraft Server",
                "players": {"max": 20, "online": 0},
                "version": {"name": "1.8-pre1", "protocol": 44}
            }
            """
            )  # overwrite `async` here
            pinger = AsyncServerPinger(
                FakeAsyncConnection(),  # type: ignore[arg-type]
                address=Address("localhost", 25565),
                version=44,
            )

            pinger.connection.receive(
                bytearray.fromhex(
                    "7200707B226465736372697074696F6E223A2241204D696E65637261667420536572766572222C22706C6179657273223A"
                    "7B226D6178223A32302C226F6E6C696E65223A307D2C2276657273696F6E223A7B226E616D65223A22312E382D70726531"
                    "222C2270726F746F636F6C223A34347D7D"
                )
            )
            # we slept 1ms, so this should be always ~1.
            assert (await pinger.read_status()).latency >= 1

    @pytest.mark.asyncio
    @pytest.mark.flaky(reruns=5, condition=sys.platform.startswith("win32"))
    async def test_test_ping_is_in_milliseconds(self):
        """``time.perf_counter`` returns fractional seconds, we must convert it to milliseconds."""

        def mocked_read_buffer():
            time.sleep(0.001)
            return mock.DEFAULT

        with mock.patch.object(FakeAsyncConnection, "read_buffer") as mocked:
            mocked.side_effect = mocked_read_buffer
            mocked.return_value.read_varint = lambda: 1  # overwrite `async` here
            mocked.return_value.read_long = lambda: 123456789  # overwrite `async` here
            pinger = AsyncServerPinger(
                FakeAsyncConnection(),  # type: ignore[arg-type]
                address=Address("localhost", 25565),
                version=44,
                ping_token=123456789,
            )
            # we slept 1ms, so this should be always ~1.
            assert await pinger.test_ping() >= 1



from __future__ import annotations

import typing

import pytest
from _pytest.python import Function, Metafunc

from tests.responses import BaseResponseTest


def pytest_generate_tests(metafunc: Metafunc) -> None:
    if issubclass(typing.cast(type, metafunc.cls), BaseResponseTest):
        instance = typing.cast(type, metafunc.cls)()
        if metafunc.definition.name not in instance._marks_table().keys():
            return

        marker_name, args = instance._marks_table()[metafunc.definition.name]
        if marker_name != "parametrize":
            return  # other markers will be handled in `pytest_collection_modifyitems`
        metafunc.parametrize(*args)


def pytest_collection_modifyitems(items: list[Function]) -> None:
    for item in items:
        if isinstance(item.instance, BaseResponseTest):
            if item.obj.__name__ not in item.instance._marks_table().keys():
                continue

            marker_name, args = item.instance._marks_table()[item.obj.__name__]
            if marker_name == "parametrize":
                continue
            item.add_marker(getattr(pytest.mark, marker_name)(*args))



from pytest import fixture

from mcstatus.motd import Motd
from mcstatus.responses import QueryPlayers, QueryResponse, QuerySoftware, RawQueryResponse
from tests.responses import BaseResponseTest


@BaseResponseTest.construct
class TestQueryResponse(BaseResponseTest):
    RAW: RawQueryResponse = RawQueryResponse(
        **{  # type: ignore # str cannot be assigned to Literal
            "hostname": "A Minecraft Server",
            "gametype": "GAME TYPE",
            "game_id": "GAME ID",
            "version": "1.8",
            "plugins": "",
            "map": "world",
            "numplayers": "3",
            "maxplayers": "20",
            "hostport": "9999",
            "hostip": "192.168.56.1",
        }
    )
    RAW_PLAYERS = ["Dinnerbone", "Djinnibone", "Steve"]

    EXPECTED_VALUES = [
        ("raw", RAW),
        ("motd", Motd.parse("A Minecraft Server")),
        ("map_name", "world"),
        ("players", QueryPlayers(online=3, max=20, list=["Dinnerbone", "Djinnibone", "Steve"])),
        ("software", QuerySoftware(version="1.8", brand="vanilla", plugins=[])),
        ("ip", "192.168.56.1"),
        ("port", 9999),
        ("game_type", "GAME TYPE"),
        ("game_id", "GAME ID"),
    ]

    @fixture(scope="class")
    def build(self):
        return QueryResponse.build(raw=self.RAW, players_list=self.RAW_PLAYERS)

    def test_as_dict(self, build: QueryResponse):
        assert build.as_dict() == {
            "game_id": "GAME ID",
            "game_type": "GAME TYPE",
            "ip": "192.168.56.1",
            "map_name": "world",
            "motd": "A Minecraft Server",
            "players": {
                "list": [
                    "Dinnerbone",
                    "Djinnibone",
                    "Steve",
                ],
                "max": 20,
                "online": 3,
            },
            "port": 9999,
            "raw": {
                "game_id": "GAME ID",
                "gametype": "GAME TYPE",
                "hostip": "192.168.56.1",
                "hostname": "A Minecraft Server",
                "hostport": "9999",
                "map": "world",
                "maxplayers": "20",
                "numplayers": "3",
                "plugins": "",
                "version": "1.8",
            },
            "software": {
                "brand": "vanilla",
                "plugins": [],
                "version": "1.8",
            },
        }


@BaseResponseTest.construct
class TestQueryPlayers(BaseResponseTest):
    EXPECTED_VALUES = [
        ("online", 3),
        ("max", 20),
        ("list", ["Dinnerbone", "Djinnibone", "Steve"]),
    ]

    @fixture(scope="class")
    def build(self):
        return QueryPlayers.build(
            raw={
                "hostname": "A Minecraft Server",
                "gametype": "SMP",
                "game_id": "MINECRAFT",
                "version": "1.8",
                "plugins": "",
                "map": "world",
                "numplayers": "3",
                "maxplayers": "20",
                "hostport": "25565",
                "hostip": "192.168.56.1",
            },
            players_list=["Dinnerbone", "Djinnibone", "Steve"],
        )


class TestQuerySoftware:
    def test_vanilla(self):
        software = QuerySoftware.build("1.8", "")
        assert software.brand == "vanilla"
        assert software.version == "1.8"
        assert software.plugins == []

    def test_modded(self):
        software = QuerySoftware.build("1.8", "A modded server: Foo 1.0; Bar 2.0; Baz 3.0")
        assert software.brand == "A modded server"
        assert software.plugins == ["Foo 1.0", "Bar 2.0", "Baz 3.0"]

    def test_modded_no_plugins(self):
        software = QuerySoftware.build("1.8", "A modded server")
        assert software.brand == "A modded server"
        assert software.plugins == []



"""Decoder for data from Forge, that is included into a response object.

After 1.18.1, Forge started to compress its mod data into a
UTF-16 string that represents binary data containing data like
the forge mod loader network version, a big list of channels
that all the forge mods use, and a list of mods the server has.

Before 1.18.1, the mod data was in `forgeData` attribute inside
a response object. We support this implementation too.

For more information see this file from forge itself:
https://github.com/MinecraftForge/MinecraftForge/blob/54b08d2711a15418130694342a3fe9a5dfe005d2/src/main/java/net/minecraftforge/network/ServerStatusPing.java#L27-L73
"""

from __future__ import annotations

from dataclasses import dataclass
from io import StringIO
from typing import Final, TYPE_CHECKING

from mcstatus.protocol.connection import BaseConnection, BaseReadSync, Connection

VERSION_FLAG_IGNORE_SERVER_ONLY: Final = 0b1
IGNORE_SERVER_ONLY: Final = "<not required for client>"


if TYPE_CHECKING:
    from typing_extensions import Self, TypedDict

    class RawForgeDataChannel(TypedDict):
        res: str
        """Channel name and ID (for example ``fml:handshake``)."""
        version: str
        """Channel version (for example ``1.2.3.4``)."""
        required: bool
        """Is this channel required for client to join?"""

    class RawForgeDataMod(TypedDict, total=False):
        modid: str
        modId: str
        modmarker: str
        """Mod version."""
        version: str

    class RawForgeData(TypedDict, total=False):
        fmlNetworkVersion: int
        channels: list[RawForgeDataChannel]
        mods: list[RawForgeDataMod]
        modList: list[RawForgeDataMod]
        d: str
        truncated: bool

else:
    RawForgeDataChannel = dict
    RawForgeDataMod = dict
    RawForgeData = dict


@dataclass(frozen=True)
class ForgeDataChannel:
    name: str
    """Channel name and ID (for example ``fml:handshake``)."""
    version: str
    """Channel version (for example ``1.2.3.4``)."""
    required: bool
    """Is this channel required for client to join?"""

    @classmethod
    def build(cls, raw: RawForgeDataChannel) -> Self:
        """Build an object about Forge channel from raw response.

        :param raw: ``channel`` element in raw forge response :class:`dict`.
        :return: :class:`ForgeDataChannel` object.
        """
        return cls(name=raw["res"], version=raw["version"], required=raw["required"])

    @classmethod
    def decode(cls, buffer: Connection, mod_id: str | None = None) -> Self:
        """Decode an object about Forge channel from decoded optimized buffer.

        :param buffer: :class:`Connection` object from UTF-16 encoded binary data.
        :param mod_id: Optional mod id prefix :class:`str`.
        :return: :class:`ForgeDataChannel` object.
        """
        channel_identifier = buffer.read_utf()
        if mod_id is not None:
            channel_identifier = f"{mod_id}:{channel_identifier}"
        version = buffer.read_utf()
        client_required = buffer.read_bool()

        return cls(
            name=channel_identifier,
            version=version,
            required=client_required,
        )


@dataclass(frozen=True)
class ForgeDataMod:
    name: str
    marker: str

    @classmethod
    def build(cls, raw: RawForgeDataMod) -> Self:
        """Build an object about Forge mod from raw response.

        :param raw: ``mod`` element in raw forge response :class:`dict`.
        :return: :class:`ForgeDataMod` object.
        """
        # In FML v1, modmarker was version instead.
        mod_version = raw.get("modmarker") or raw.get("version")
        if mod_version is None:
            raise KeyError("Mod version in Forge mod data must be provided.")

        # In FML v2, modid was modId instead. At least one of the two should exist.
        mod_id = raw.get("modid") or raw.get("modId")
        if mod_id is None:
            raise KeyError(f"Mod ID in Forge mod data must be provided. Mod version: {mod_version!r}.")

        return cls(name=mod_id, marker=mod_version)

    @classmethod
    def decode(cls, buffer: Connection) -> tuple[Self, list[ForgeDataChannel]]:
        """Decode data about a Forge mod from decoded optimized buffer.

        :param buffer: :class:`Connection` object from UTF-16 encoded binary data.
        :return: :class:`tuple` object of :class:`ForgeDataMod` object and :class:`list` of :class:`ForgeDataChannel` objects.
        """
        channel_version_flags = buffer.read_varint()

        channel_count = channel_version_flags >> 1
        is_server = channel_version_flags & VERSION_FLAG_IGNORE_SERVER_ONLY != 0
        mod_id = buffer.read_utf()

        mod_version = IGNORE_SERVER_ONLY
        if not is_server:
            mod_version = buffer.read_utf()

        channels: list[ForgeDataChannel] = []
        for _ in range(channel_count):
            channels.append(ForgeDataChannel.decode(buffer, mod_id))

        return cls(name=mod_id, marker=mod_version), channels


class StringBuffer(BaseReadSync, BaseConnection):
    """String Buffer for reading utf-16 encoded binary data."""

    __slots__ = ("received", "stringio")

    def __init__(self, stringio: StringIO) -> None:
        self.stringio = stringio
        self.received = bytearray()

    def read(self, length: int) -> bytearray:
        """Read length bytes from ``self``, and return a byte array."""
        data = bytearray()
        while self.received and len(data) < length:
            data.append(self.received.pop(0))
        while len(data) < length:
            result = self.stringio.read(1)
            if not result:
                raise IOError(f"Not enough data to read! {len(data)} < {length}")
            data.extend(result.encode("utf-16be"))
        while len(data) > length:
            self.received.append(data.pop())
        return data

    def remaining(self) -> int:
        """Return number of reads remaining."""
        return len(self.stringio.getvalue()) - self.stringio.tell() + len(self.received)

    def read_optimized_size(self) -> int:
        """Read encoded data length."""
        return self.read_short() | (self.read_short() << 15)

    def read_optimized_buffer(self) -> Connection:
        """Read encoded buffer."""
        size = self.read_optimized_size()

        buffer = Connection()
        value, bits = 0, 0
        while buffer.remaining() < size:
            if bits < 8 and self.remaining():
                # Ignoring sign bit
                value |= (self.read_short() & 0x7FFF) << bits
                bits += 15
            buffer.receive((value & 0xFF).to_bytes(1, "big"))
            value >>= 8
            bits -= 8

        return buffer


@dataclass(frozen=True)
class ForgeData:
    fml_network_version: int
    """Forge Mod Loader network version."""
    channels: list[ForgeDataChannel]
    """List of channels, both for mods and non-mods."""
    mods: list[ForgeDataMod]
    """List of mods"""
    truncated: bool
    """Is the mods list and or channel list incomplete?"""

    @staticmethod
    def _decode_optimized(string: str) -> Connection:
        """Decode buffer from UTF-16 optimized binary data ``string``."""
        with StringIO(string) as text:
            str_buffer = StringBuffer(text)
            return str_buffer.read_optimized_buffer()

    @classmethod
    def build(cls, raw: RawForgeData) -> Self | None:
        """Build an object about Forge mods from raw response.

        :param raw: ``forgeData`` attribute in raw response :class:`dict`.
        :return: :class:`ForgeData` object.
        """
        fml_network_version = raw.get("fmlNetworkVersion", 1)

        # see https://github.com/MinecraftForge/MinecraftForge/blob/7d0330eb08299935714e34ac651a293e2609aa86/src/main/java/net/minecraftforge/network/ServerStatusPing.java#L27-L73  # noqa: E501  # line too long
        if "d" not in raw:
            mod_list = raw.get("mods") or raw.get("modList")
            if mod_list is None:
                raise KeyError("Neither `mods` or `modList` keys exist.")

            return cls(
                fml_network_version=fml_network_version,
                channels=[ForgeDataChannel.build(channel) for channel in raw.get("channels", ())],
                mods=[ForgeDataMod.build(mod) for mod in mod_list],
                truncated=False,
            )

        buffer = cls._decode_optimized(raw["d"])

        channels: list[ForgeDataChannel] = []
        mods: list[ForgeDataMod] = []

        truncated = buffer.read_bool()
        mod_count = buffer.read_ushort()
        try:
            for _ in range(mod_count):
                mod, mod_channels = ForgeDataMod.decode(buffer)

                channels.extend(mod_channels)
                mods.append(mod)

            non_mod_channel_count = buffer.read_varint()
            for _ in range(non_mod_channel_count):
                channels.append(ForgeDataChannel.decode(buffer))
        except IOError:
            if not truncated:
                raise  # If answer wasn't truncated, we lost some data on the way

        return cls(
            fml_network_version=fml_network_version,
            channels=channels,
            mods=mods,
            truncated=truncated,
        )



from __future__ import annotations
import asyncio

from mcstatus import BedrockServer, JavaServer
from mcstatus.responses import BedrockStatusResponse, JavaStatusResponse


async def status(host: str) -> JavaStatusResponse | BedrockStatusResponse:
    """Get status from server, which can be Java or Bedrock.

    The function will ping server as Java and as Bedrock in one time, and return the first response.
    """
    success_task = await handle_exceptions(
        *(
            await asyncio.wait(
                {
                    asyncio.create_task(handle_java(host), name="Get status as Java"),
                    asyncio.create_task(handle_bedrock(host), name="Get status as Bedrock"),
                },
                return_when=asyncio.FIRST_COMPLETED,
            )
        )
    )

    if success_task is None:
        raise ValueError("No tasks were successful. Is server offline?")

    return success_task.result()


async def handle_exceptions(done: set[asyncio.Task], pending: set[asyncio.Task]) -> asyncio.Task | None:
    """Handle exceptions from tasks.

    Also, cancel all pending tasks, if found correct one.
    """
    if len(done) == 0:
        raise ValueError("No tasks was given to `done` set.")

    for i, task in enumerate(done):
        if task.exception() is not None:
            if len(pending) == 0:
                continue

            if i == len(done) - 1:  # firstly check all items from `done` set, and then handle pending set
                return await handle_exceptions(*(await asyncio.wait(pending, return_when=asyncio.FIRST_COMPLETED)))
        else:
            for pending_task in pending:
                pending_task.cancel()
            return task


async def handle_java(host: str) -> JavaStatusResponse:
    """A wrapper around mcstatus, to compress it in one function."""
    return await (await JavaServer.async_lookup(host)).async_status()


async def handle_bedrock(host: str) -> BedrockStatusResponse:
    """A wrapper around mcstatus, to compress it in one function."""
    # note: `BedrockServer` doesn't have `async_lookup` method, see it's docstring
    return await BedrockServer.lookup(host).async_status()



from unittest.mock import Mock, patch

import pytest

from mcstatus.address import Address
from mcstatus.protocol.connection import Connection, TCPSocketConnection, UDPSocketConnection


class TestConnection:
    connection: Connection

    def setup_method(self):
        self.connection = Connection()

    def test_flush(self):
        self.connection.sent = bytearray.fromhex("7FAABB")

        assert self.connection.flush() == bytearray.fromhex("7FAABB")
        assert self.connection.sent == bytearray()

    def test_receive(self):
        self.connection.receive(bytearray.fromhex("7F"))
        self.connection.receive(bytearray.fromhex("AABB"))

        assert self.connection.received == bytearray.fromhex("7FAABB")

    def test_remaining(self):
        self.connection.receive(bytearray.fromhex("7F"))
        self.connection.receive(bytearray.fromhex("AABB"))

        assert self.connection.remaining() == 3

    def test_send(self):
        self.connection.write(bytearray.fromhex("7F"))
        self.connection.write(bytearray.fromhex("AABB"))

        assert self.connection.flush() == bytearray.fromhex("7FAABB")

    def test_read(self):
        self.connection.receive(bytearray.fromhex("7FAABB"))

        assert self.connection.read(2) == bytearray.fromhex("7FAA")
        assert self.connection.read(1) == bytearray.fromhex("BB")

    def _assert_varint_read_write(self, hexstr, value) -> None:
        self.connection.receive(bytearray.fromhex(hexstr))
        assert self.connection.read_varint() == value

        self.connection.write_varint(value)
        assert self.connection.flush() == bytearray.fromhex(hexstr)

    def test_varint_cases(self):
        self._assert_varint_read_write("00", 0)
        self._assert_varint_read_write("01", 1)
        self._assert_varint_read_write("0F", 15)
        self._assert_varint_read_write("FFFFFFFF07", 2147483647)

        self._assert_varint_read_write("FFFFFFFF0F", -1)
        self._assert_varint_read_write("8080808008", -2147483648)

    def test_read_invalid_varint(self):
        self.connection.receive(bytearray.fromhex("FFFFFFFF80"))

        with pytest.raises(IOError):
            self.connection.read_varint()

    def test_write_invalid_varint(self):
        with pytest.raises(ValueError):
            self.connection.write_varint(2147483648)
        with pytest.raises(ValueError):
            self.connection.write_varint(-2147483649)

    def test_read_utf(self):
        self.connection.receive(bytearray.fromhex("0D48656C6C6F2C20776F726C6421"))

        assert self.connection.read_utf() == "Hello, world!"

    def test_write_utf(self):
        self.connection.write_utf("Hello, world!")

        assert self.connection.flush() == bytearray.fromhex("0D48656C6C6F2C20776F726C6421")

    def test_read_empty_utf(self):
        self.connection.write_utf("")

        assert self.connection.flush() == bytearray.fromhex("00")

    def test_read_ascii(self):
        self.connection.receive(bytearray.fromhex("48656C6C6F2C20776F726C642100"))

        assert self.connection.read_ascii() == "Hello, world!"

    def test_write_ascii(self):
        self.connection.write_ascii("Hello, world!")

        assert self.connection.flush() == bytearray.fromhex("48656C6C6F2C20776F726C642100")

    def test_read_empty_ascii(self):
        self.connection.write_ascii("")

        assert self.connection.flush() == bytearray.fromhex("00")

    def test_read_short_negative(self):
        self.connection.receive(bytearray.fromhex("8000"))

        assert self.connection.read_short() == -32768

    def test_write_short_negative(self):
        self.connection.write_short(-32768)

        assert self.connection.flush() == bytearray.fromhex("8000")

    def test_read_short_positive(self):
        self.connection.receive(bytearray.fromhex("7FFF"))

        assert self.connection.read_short() == 32767

    def test_write_short_positive(self):
        self.connection.write_short(32767)

        assert self.connection.flush() == bytearray.fromhex("7FFF")

    def test_read_ushort_positive(self):
        self.connection.receive(bytearray.fromhex("8000"))

        assert self.connection.read_ushort() == 32768

    def test_write_ushort_positive(self):
        self.connection.write_ushort(32768)

        assert self.connection.flush() == bytearray.fromhex("8000")

    def test_read_int_negative(self):
        self.connection.receive(bytearray.fromhex("80000000"))

        assert self.connection.read_int() == -2147483648

    def test_write_int_negative(self):
        self.connection.write_int(-2147483648)

        assert self.connection.flush() == bytearray.fromhex("80000000")

    def test_read_int_positive(self):
        self.connection.receive(bytearray.fromhex("7FFFFFFF"))

        assert self.connection.read_int() == 2147483647

    def test_write_int_positive(self):
        self.connection.write_int(2147483647)

        assert self.connection.flush() == bytearray.fromhex("7FFFFFFF")

    def test_read_uint_positive(self):
        self.connection.receive(bytearray.fromhex("80000000"))

        assert self.connection.read_uint() == 2147483648

    def test_write_uint_positive(self):
        self.connection.write_uint(2147483648)

        assert self.connection.flush() == bytearray.fromhex("80000000")

    def test_read_long_negative(self):
        self.connection.receive(bytearray.fromhex("8000000000000000"))

        assert self.connection.read_long() == -9223372036854775808

    def test_write_long_negative(self):
        self.connection.write_long(-9223372036854775808)

        assert self.connection.flush() == bytearray.fromhex("8000000000000000")

    def test_read_long_positive(self):
        self.connection.receive(bytearray.fromhex("7FFFFFFFFFFFFFFF"))

        assert self.connection.read_long() == 9223372036854775807

    def test_write_long_positive(self):
        self.connection.write_long(9223372036854775807)

        assert self.connection.flush() == bytearray.fromhex("7FFFFFFFFFFFFFFF")

    def test_read_ulong_positive(self):
        self.connection.receive(bytearray.fromhex("8000000000000000"))

        assert self.connection.read_ulong() == 9223372036854775808

    def test_write_ulong_positive(self):
        self.connection.write_ulong(9223372036854775808)

        assert self.connection.flush() == bytearray.fromhex("8000000000000000")

    @pytest.mark.parametrize("as_bytes,as_bool", [("01", True), ("00", False)])
    def test_read_bool(self, as_bytes: str, as_bool: bool) -> None:
        self.connection.receive(bytearray.fromhex(as_bytes))

        assert self.connection.read_bool() is as_bool

    @pytest.mark.parametrize("as_bytes,as_bool", [("01", True), ("00", False)])
    def test_write_bool(self, as_bytes: str, as_bool: bool) -> None:
        self.connection.write_bool(as_bool)

        assert self.connection.flush() == bytearray.fromhex(as_bytes)

    def test_read_buffer(self):
        self.connection.receive(bytearray.fromhex("027FAA"))
        buffer = self.connection.read_buffer()

        assert buffer.received == bytearray.fromhex("7FAA")
        assert self.connection.flush() == bytearray()

    def test_write_buffer(self):
        buffer = Connection()
        buffer.write(bytearray.fromhex("7FAA"))
        self.connection.write_buffer(buffer)

        assert self.connection.flush() == bytearray.fromhex("027FAA")

    def test_read_empty(self):
        self.connection.received = bytearray()

        with pytest.raises(IOError):
            self.connection.read(1)

    def test_read_not_enough(self):
        self.connection.received = bytearray(b"a")

        with pytest.raises(IOError):
            self.connection.read(2)


class TestTCPSocketConnection:
    @pytest.fixture(scope="class")
    def connection(self):
        test_addr = Address("localhost", 1234)

        socket = Mock()
        socket.recv = Mock()
        socket.send = Mock()
        with patch("socket.create_connection") as create_connection:
            create_connection.return_value = socket
            with TCPSocketConnection(test_addr) as connection:
                yield connection

    def test_flush(self, connection):
        with pytest.raises(TypeError):
            connection.flush()

    def test_receive(self, connection):
        with pytest.raises(TypeError):
            connection.receive("")  # type: ignore # This is desired to produce TypeError

    def test_remaining(self, connection):
        with pytest.raises(TypeError):
            connection.remaining()

    def test_read(self, connection):
        connection.socket.recv.return_value = bytearray.fromhex("7FAA")

        assert connection.read(2) == bytearray.fromhex("7FAA")

    def test_read_empty(self, connection):
        connection.socket.recv.return_value = bytearray()

        with pytest.raises(IOError):
            connection.read(1)

    def test_read_not_enough(self, connection):
        connection.socket.recv.side_effect = [bytearray(b"a"), bytearray()]

        with pytest.raises(IOError):
            connection.read(2)

    def test_write(self, connection):
        connection.write(bytearray.fromhex("7FAA"))

        connection.socket.send.assert_called_once_with(bytearray.fromhex("7FAA"))  # type: ignore[attr-defined]


class TestUDPSocketConnection:
    @pytest.fixture(scope="class")
    def connection(self):
        test_addr = Address("localhost", 1234)

        socket = Mock()
        socket.recvfrom = Mock()
        socket.sendto = Mock()
        with patch("socket.socket") as create_socket:
            create_socket.return_value = socket
            with UDPSocketConnection(test_addr) as connection:
                yield connection

    def test_flush(self, connection):
        with pytest.raises(TypeError):
            connection.flush()

    def test_receive(self, connection):
        with pytest.raises(TypeError):
            connection.receive("")  # type: ignore # This is desired to produce TypeError

    def test_remaining(self, connection):
        assert connection.remaining() == 65535

    def test_read(self, connection):
        connection.socket.recvfrom.return_value = [bytearray.fromhex("7FAA")]

        assert connection.read(2) == bytearray.fromhex("7FAA")

    def test_read_empty(self, connection):
        connection.socket.recvfrom.return_value = []

        with pytest.raises(IndexError):
            connection.read(1)

    def test_write(self, connection):
        connection.write(bytearray.fromhex("7FAA"))

        connection.socket.sendto.assert_called_once_with(  # type: ignore[attr-defined]
            bytearray.fromhex("7FAA"),
            Address("localhost", 1234),
        )



from __future__ import annotations

import abc
from typing import Any, TypeVar, cast

from mcstatus.responses import BaseStatusResponse

__all__ = ["BaseResponseTest"]
_T = TypeVar("_T", bound="type[BaseResponseTest]")


class BaseResponseTest(abc.ABC):
    EXPECTED_VALUES: list[tuple[str, Any]] | None = None
    EXPECTED_TYPES: list[tuple[str, type]] | None = None
    ATTRIBUTES_IN: list[str] | None = None
    # if we don't specify item in raw answer, target field will be None
    # a first element is a list with fields to remove, and attribute that
    # must be None. a dict is a raw answer to pass into `build` method
    OPTIONAL_FIELDS: tuple[list[tuple[str, str]], dict[str, Any]] | None = None

    def _validate(self) -> None:
        """Perform checks to validate the class."""
        if self.EXPECTED_TYPES is not None and self.EXPECTED_VALUES is not None:
            expected_values_keys = list(dict(self.EXPECTED_VALUES).keys())

            for key in dict(self.EXPECTED_TYPES).keys():
                if key in expected_values_keys:
                    raise ValueError("You can't test the type of attribute, if already testing its value.")

        if self.ATTRIBUTES_IN is not None and (self.EXPECTED_VALUES is not None or self.EXPECTED_TYPES is not None):
            if self.EXPECTED_VALUES and self.EXPECTED_TYPES:
                to_dict = self.EXPECTED_VALUES.copy()
                to_dict.extend(self.EXPECTED_TYPES)
                already_checked_attributes = dict(to_dict).keys()
            else:
                already_checked_attributes = dict(self.EXPECTED_VALUES or self.EXPECTED_TYPES).keys()  # type: ignore

            for attribute_name in self.ATTRIBUTES_IN:
                if attribute_name in already_checked_attributes:
                    raise ValueError("You can't test the type availability, if already testing its value/type.")

    @abc.abstractmethod
    def build(self) -> Any:  # noqa: ANN401
        ...

    # implementations for tests

    def test_values_of_attributes(self, build: BaseStatusResponse, field: str, value: Any) -> None:  # noqa: ANN401
        assert getattr(build, field) == value

    def test_types_of_attributes(self, build: BaseStatusResponse, field: str, type_: type) -> None:
        assert isinstance(getattr(build, field), type_)

    def test_attribute_in(self, build: BaseStatusResponse, field: str) -> None:
        assert hasattr(build, field)

    def test_optional_field_turns_into_none(self, build: BaseStatusResponse, to_remove: str, attribute_name: str) -> None:
        raw = cast(tuple, self.OPTIONAL_FIELDS)[1]
        del raw[to_remove]
        assert getattr(type(build).build(raw), attribute_name) is None  # type: ignore # build is abstract

    def _dependency_table(self) -> dict[str, bool]:
        # a key in the dict must be a name of a test implementation.
        # and a value of the dict is a bool. if it's false - we
        # "delete" a test from the class.
        return {
            "test_values_of_attributes": self.EXPECTED_VALUES is not None,
            "test_types_of_attributes": self.EXPECTED_TYPES is not None,
            "test_attribute_in": self.ATTRIBUTES_IN is not None,
            "test_optional_field_turns_into_none": self.OPTIONAL_FIELDS is not None,
        }

    def _marks_table(self) -> dict[str, tuple[str, tuple[Any, ...]]]:
        # hooks in conftest.py parses this table

        # a key in the dict must be a name of a test implementation.
        # and a value of the dict is a tuple, where first element is
        # a name of mark to apply to the test, and second element is
        # positional arguments, which passed to the mark
        return {
            "test_values_of_attributes": ("parametrize", ("field,value", self.EXPECTED_VALUES)),
            "test_types_of_attributes": ("parametrize", ("field,type_", self.EXPECTED_TYPES)),
            "test_attribute_in": ("parametrize", ("field", self.ATTRIBUTES_IN)),
            "test_optional_field_turns_into_none": (
                "parametrize",
                ("to_remove,attribute_name", self.OPTIONAL_FIELDS[0] if self.OPTIONAL_FIELDS is not None else ()),
            ),
        }

    @staticmethod
    def construct(class_: _T) -> _T:
        instance: BaseResponseTest = class_()  # type: ignore
        instance._validate()
        for implementation_name, meet_dependencies in instance._dependency_table().items():
            if not meet_dependencies:
                # delattr works only with initialized classes,
                # hopefully overwriting with None doesn't have this limitation
                setattr(class_, implementation_name, None)

        return class_



from mcstatus.responses import (
    BaseStatusPlayers,
    BaseStatusResponse,
    BaseStatusVersion,
    BedrockStatusPlayers,
    BedrockStatusResponse,
    BedrockStatusVersion,
    JavaStatusPlayer,
    JavaStatusPlayers,
    JavaStatusResponse,
    JavaStatusVersion,
)

# __all__ is frozen on the moment of deprecation
__all__ = [
    "BaseStatusPlayers",
    "BaseStatusResponse",
    "BaseStatusVersion",
    "BedrockStatusPlayers",
    "BedrockStatusResponse",
    "BedrockStatusVersion",
    "JavaStatusPlayer",
    "JavaStatusPlayers",
    "JavaStatusResponse",
    "JavaStatusVersion",
]


import warnings

warnings.warn(
    "`mcstatus.status_response` is deprecated, and will be removed at 2025-12, use `mcstatus.responses` instead",
    DeprecationWarning,
    stacklevel=2,
)



import pytest

from mcstatus.forge_data import ForgeData, ForgeDataChannel, ForgeDataMod, RawForgeData
from tests.responses import BaseResponseTest


@BaseResponseTest.construct
class TestForgeDataV1(BaseResponseTest):
    RAW = {
        "type": "FML",
        "modList": [
            {"modid": "minecraft", "version": "1.12.2"},
            {"modid": "mcp", "version": "9.42"},
            {"modid": "FML", "version": "8.0.99.99"},
            {"modid": "forge", "version": "14.23.5.2859"},
        ],
        "channels": [
            {"res": "fml:handshake", "version": "1.2.3.4", "required": True},
        ],
    }

    EXPECTED_VALUES = [
        ("fml_network_version", 1),
        ("channels", [ForgeDataChannel(name="fml:handshake", version="1.2.3.4", required=True)]),
        (
            "mods",
            [
                ForgeDataMod(name="minecraft", marker="1.12.2"),
                ForgeDataMod(name="mcp", marker="9.42"),
                ForgeDataMod(name="FML", marker="8.0.99.99"),
                ForgeDataMod(name="forge", marker="14.23.5.2859"),
            ],
        ),
        ("truncated", False),
    ]

    @pytest.fixture(scope="class")
    def build(self) -> ForgeData:
        return ForgeData.build(self.RAW)  # type: ignore # dict[str, Unknown] cannot be assigned to TypedDict


@BaseResponseTest.construct
class TestForgeDataV2(BaseResponseTest):
    RAW = {
        "fmlNetworkVersion": 2,
        "channels": [
            {"res": "fml:handshake", "version": "1.2.3.4", "required": True},
        ],
        "mods": [
            {"modId": "forge", "modmarker": "ANY"},
        ],
    }

    EXPECTED_VALUES = [
        ("fml_network_version", 2),
        ("channels", [ForgeDataChannel(name="fml:handshake", version="1.2.3.4", required=True)]),
        ("mods", [ForgeDataMod(name="forge", marker="ANY")]),
        ("truncated", False),
    ]

    @pytest.fixture(scope="class")
    def build(self) -> ForgeData:
        return ForgeData.build(self.RAW)  # type: ignore # dict[str, Unknown] cannot be assigned to TypedDict


@BaseResponseTest.construct
class TestForgeDataV3(BaseResponseTest):
    RAW = {
        "channels": [],
        "mods": [],
        "truncated": False,
        "fmlNetworkVersion": 3,
        "d": bytes.fromhex(
            "5e0000e0a084e390a4e78d8be39996e2b98ce1a698ccbae2b8b1e681a4e492b8e2a191e29ba7e6b2aee5a"
            "999e3a8b9e789a5e0b088e384b5e0a69ae28280e6b2aee5a999e3a8b9e789a5e0b088e384b5e0a69ae581"
            "80e6b380e5b29be38ab3e48483e38a9ce580b1e2ad8be79ca6e6b9abe1b29be392bae69daee68886e482b"
            "8e2a081dcb0e2b68ee5b49ae1a281e384ae02"
        ).decode("utf8"),
    }

    EXPECTED_VALUES = [
        ("fml_network_version", 3),
        (
            "channels",
            [
                ForgeDataChannel(name="minecraft:unregister", version="FML3", required=True),
                ForgeDataChannel(name="minecraft:register", version="FML3", required=True),
                ForgeDataChannel(name="forge:tier_sorting", version="1.0", required=False),
                ForgeDataChannel(name="forge:split", version="1.1", required=True),
            ],
        ),
        (
            "mods",
            [
                ForgeDataMod(name="minecraft", marker="1.20.1"),
                ForgeDataMod(name="forge", marker="ANY"),
            ],
        ),
        ("truncated", False),
    ]

    @pytest.fixture(scope="class")
    def build(self) -> ForgeData:
        return ForgeData.build(self.RAW)  # type: ignore # dict[str, Unknown] cannot be assigned to TypedDict


@BaseResponseTest.construct
class TestForgeData(BaseResponseTest):
    EXPECTED_VALUES = [
        ("fml_network_version", 3),
        (
            "channels",
            [
                ForgeDataChannel(
                    name="cyclopscore:channel_main",
                    version="1.0.0",
                    required=True,
                ),
                ForgeDataChannel(
                    name="supermartijn642configlib:sync_configs",
                    version="1",
                    required=False,
                ),
                ForgeDataChannel(
                    name="alexsmobs:main_channel",
                    version="1",
                    required=False,
                ),
                ForgeDataChannel(
                    name="sophisticatedcore:channel",
                    version="1",
                    required=False,
                ),
                ForgeDataChannel(
                    name="rftoolsbase:rftoolsbase",
                    version="1.0",
                    required=True,
                ),
                ForgeDataChannel(
                    name="irongenerators:irongenerators",
                    version="1",
                    required=False,
                ),
                ForgeDataChannel(
                    name="xaeroworldmap:main",
                    version="1.0",
                    required=True,
                ),
                ForgeDataChannel(
                    name="cookingforblockheads:network",
                    version="1.0",
                    required=False,
                ),
                ForgeDataChannel(name="xnet:xnet", version="1.0", required=True),
                ForgeDataChannel(
                    name="placebo:placebo",
                    version="1.0.0",
                    required=True,
                ),
                ForgeDataChannel(
                    name="citadel:main_channel",
                    version="1",
                    required=False,
                ),
                ForgeDataChannel(
                    name="sophisticatedbackpacks:channel",
                    version="1",
                    required=False,
                ),
                ForgeDataChannel(
                    name="buildinggadgets:main",
                    version="4",
                    required=False,
                ),
                ForgeDataChannel(
                    name="mekanismgenerators:mekanismgenerators",
                    version="10.2.5",
                    required=False,
                ),
                ForgeDataChannel(
                    name="waila:networking",
                    version="1.0.0",
                    required=True,
                ),
                ForgeDataChannel(
                    name="shetiphiancore:main_channel",
                    version="1.0.0",
                    required=False,
                ),
                ForgeDataChannel(
                    name="dummmmmmy:dummychannel",
                    version="1",
                    required=False,
                ),
                ForgeDataChannel(
                    name="supplementaries:network",
                    version="1",
                    required=False,
                ),
                ForgeDataChannel(
                    name="refinedstorage:main_channel",
                    version="1",
                    required=False,
                ),
                ForgeDataChannel(name="corpse:default", version="1.0.0", required=True),
                ForgeDataChannel(
                    name="ping:ping_channel",
                    version="PING1",
                    required=True,
                ),
                ForgeDataChannel(
                    name="ironfurnaces:ironfurnaces_network",
                    version="1.0",
                    required=True,
                ),
                ForgeDataChannel(name="botania:main", version="0", required=False),
                ForgeDataChannel(name="curios:main", version="1", required=False),
                ForgeDataChannel(name="patchouli:main", version="1", required=False),
                ForgeDataChannel(name="camera:default", version="1.0.0", required=True),
                ForgeDataChannel(
                    name="libnonymous:channel",
                    version="1.0",
                    required=True,
                ),
                ForgeDataChannel(
                    name="elevatorid:main_channel",
                    version="1",
                    required=False,
                ),
                ForgeDataChannel(name="worldedit:cui", version="1", required=True),
                ForgeDataChannel(name="worldedit:internal", version="1", required=True),
                ForgeDataChannel(name="cfm:network", version="1", required=False),
                ForgeDataChannel(
                    name="architectury:network",
                    version="1",
                    required=True,
                ),
                ForgeDataChannel(name="trashcans:main", version="1", required=False),
                ForgeDataChannel(name="jei:channel", version="1.0.0", required=True),
                ForgeDataChannel(name="ae2:main", version="1", required=True),
                ForgeDataChannel(
                    name="mekanism:mekanism",
                    version="10.2.5",
                    required=False,
                ),
                ForgeDataChannel(name="bdlib:multiblock", version="2", required=False),
                ForgeDataChannel(name="bdlib:misc", version="1", required=False),
                ForgeDataChannel(name="create:main", version="1", required=False),
                ForgeDataChannel(
                    name="waystones:network",
                    version="1.0",
                    required=False,
                ),
                ForgeDataChannel(name="comforts:main", version="1", required=False),
                ForgeDataChannel(
                    name="naturescompass:naturescompass",
                    version="1.0",
                    required=True,
                ),
                ForgeDataChannel(
                    name="storagenetwork:main_channel",
                    version="1",
                    required=False,
                ),
                ForgeDataChannel(name="cofh_core:general", version="1", required=True),
                ForgeDataChannel(
                    name="mcjtylib:mcjtylib",
                    version="1.0",
                    required=True,
                ),
                ForgeDataChannel(
                    name="mininggadgets:main_network_channel",
                    version="2",
                    required=False,
                ),
                ForgeDataChannel(
                    name="crafttweaker:main",
                    version="1.0.0",
                    required=False,
                ),
                ForgeDataChannel(name="akashictome:main", version="1", required=False),
                ForgeDataChannel(
                    name="forge:tier_sorting",
                    version="1.0",
                    required=False,
                ),
                ForgeDataChannel(name="forge:split", version="1.1", required=True),
                ForgeDataChannel(
                    name="colossalchests:channel_main",
                    version="1.0.0",
                    required=True,
                ),
                ForgeDataChannel(name="selene:network", version="1", required=False),
                ForgeDataChannel(
                    name="craftingtweaks:network",
                    version="1.0",
                    required=False,
                ),
                ForgeDataChannel(
                    name="minecraft:unregister",
                    version="FML3",
                    required=True,
                ),
                ForgeDataChannel(
                    name="minecraft:register",
                    version="FML3",
                    required=True,
                ),
                ForgeDataChannel(name="titanium:network", version="1.0", required=True),
                ForgeDataChannel(
                    name="easy_villagers:default",
                    version="1.0.0",
                    required=True,
                ),
                ForgeDataChannel(name="pipez:default", version="1.0.0", required=True),
                ForgeDataChannel(name="mantle:network", version="1", required=False),
                ForgeDataChannel(name="quark:main", version="1", required=False),
                ForgeDataChannel(
                    name="xaerominimap:main",
                    version="1.0",
                    required=True,
                ),
                ForgeDataChannel(
                    name="fastbench:channel",
                    version="4.6.0",
                    required=True,
                ),
                ForgeDataChannel(name="polymorph:main", version="1", required=False),
                ForgeDataChannel(
                    name="storagedrawers:main_channel",
                    version="1",
                    required=False,
                ),
                ForgeDataChannel(
                    name="enercell:network",
                    version="0.0.0",
                    required=False,
                ),
                ForgeDataChannel(name="appleskin:sync", version="1", required=True),
                ForgeDataChannel(
                    name="modularrouters:main_channel",
                    version="2",
                    required=False,
                ),
                ForgeDataChannel(
                    name="the_vault:network",
                    version="0.26.0",
                    required=False,
                ),
                ForgeDataChannel(
                    name="modernui:fluxnetworks",
                    version="707",
                    required=False,
                ),
            ],
        ),
        (
            "mods",
            [
                ForgeDataMod(name="rsrequestify", marker="2.2.0"),
                ForgeDataMod(name="cyclopscore", marker="1.15.1"),
                ForgeDataMod(name="auudio", marker="1.0.3"),
                ForgeDataMod(name="auxiliaryblocks", marker="1.18.2-0.0.14"),
                ForgeDataMod(name="supermartijn642configlib", marker="1.1.6"),
                ForgeDataMod(name="alexsmobs", marker="1.18.6"),
                ForgeDataMod(name="architects_palette", marker="1.1.2"),
                ForgeDataMod(name="cagerium", marker="1.18.2-1.1.0"),
                ForgeDataMod(name="mcwwindows", marker="2.0.3"),
                ForgeDataMod(
                    name="sophisticatedcore",
                    marker="1.18.2-0.5.32.179",
                ),
                ForgeDataMod(name="thermal", marker="1.6.3.28"),
                ForgeDataMod(name="rftoolsbase", marker="1.18-3.0.9"),
                ForgeDataMod(name="initialinventory", marker="6.0.8"),
                ForgeDataMod(name="irongenerators", marker="2.0.1"),
                ForgeDataMod(name="xaeroworldmap", marker="1.25.1"),
                ForgeDataMod(name="cookingforblockheads", marker="12.0.2"),
                ForgeDataMod(
                    name="controlling",
                    marker="<not required for client>",
                ),
                ForgeDataMod(name="xnet", marker="1.18-4.0.5"),
                ForgeDataMod(name="placebo", marker="6.4.1"),
                ForgeDataMod(name="citadel", marker="1.11.3"),
                ForgeDataMod(name="powah", marker="3.0.1-beta"),
                ForgeDataMod(name="bookshelf", marker="13.2.50"),
                ForgeDataMod(name="lootbeams", marker="1.18.1"),
                ForgeDataMod(
                    name="sophisticatedbackpacks",
                    marker="1.18.2-3.18.35.752",
                ),
                ForgeDataMod(name="twigs", marker="1.1.4-patch1+1.18.2"),
                ForgeDataMod(
                    name="buildinggadgets",
                    marker="3.13.0-build.5+mc1.18.2}",
                ),
                ForgeDataMod(name="darkutils", marker="10.0.5"),
                ForgeDataMod(name="mcwdoors", marker="1.0.6"),
                ForgeDataMod(name="waddles", marker="1.18.2-0.8.19"),
                ForgeDataMod(name="mekanismgenerators", marker="10.2.5"),
                ForgeDataMod(name="balm", marker="3.2.0+0"),
                ForgeDataMod(name="waila", marker="<not required for client>"),
                ForgeDataMod(name="jeresources", marker="0.14.1.171"),
                ForgeDataMod(
                    name="cloth_config",
                    marker="<not required for client>",
                ),
                ForgeDataMod(name="shetiphiancore", marker="3.10.10"),
                ForgeDataMod(name="dummmmmmy", marker="1.18-1.5.2"),
                ForgeDataMod(name="supplementaries", marker="1.18.2-1.5.13"),
                ForgeDataMod(name="refinedstorage", marker="1.10.2"),
                ForgeDataMod(name="konkrete", marker="1.3.3"),
                ForgeDataMod(name="easy_piglins", marker="1.18.2-1.0.0"),
                ForgeDataMod(name="corpse", marker="1.18.2-1.0.2"),
                ForgeDataMod(name="packmenu", marker="<not required for client>"),
                ForgeDataMod(name="mcwbridges", marker="2.0.3"),
                ForgeDataMod(name="torchmaster", marker="18.1.0"),
                ForgeDataMod(name="compressium", marker="1.4.2-build.9+mc1.18.2"),
                ForgeDataMod(name="ping", marker="1.18-1.8.0"),
                ForgeDataMod(name="ironfurnaces", marker="3.3.1"),
                ForgeDataMod(name="mcwtrpdoors", marker="1.0.6"),
                ForgeDataMod(name="mcwfences", marker="1.0.5"),
                ForgeDataMod(name="supermartijn642corelib", marker="1.0.19"),
                ForgeDataMod(name="simplylight", marker="1.18.2-1.4.2-build.31"),
                ForgeDataMod(name="botania", marker="1.18.2-434"),
                ForgeDataMod(name="highlighter", marker="ANY"),
                ForgeDataMod(name="spark", marker="<not required for client>"),
                ForgeDataMod(name="curios", marker="1.18.2-5.0.7.1"),
                ForgeDataMod(name="patchouli", marker="1.18.2-71.1"),
                ForgeDataMod(name="camera", marker="1.18.2-1.0.4"),
                ForgeDataMod(name="blockcarpentry", marker="1.18-0.3.0"),
                ForgeDataMod(name="thermal_foundation", marker="1.6.3.28"),
                ForgeDataMod(name="thermal_expansion", marker="1.6.3.13"),
                ForgeDataMod(name="libnonymous", marker="2.1.0"),
                ForgeDataMod(name="elevatorid", marker="1.18.2-1.8.4"),
                ForgeDataMod(name="runelic", marker="11.0.1"),
                ForgeDataMod(
                    name="worldedit",
                    marker="<not required for client>",
                ),
                ForgeDataMod(name="cfm", marker="7.0.0-pre29"),
                ForgeDataMod(name="architectury", marker="4.9.84"),
                ForgeDataMod(name="weirdinggadget", marker="2.2.11"),
                ForgeDataMod(name="mcwfurnitures", marker="3.0.0"),
                ForgeDataMod(name="trashcans", marker="1.0.15"),
                ForgeDataMod(name="mcwlights", marker="1.0.3"),
                ForgeDataMod(name="cucumber", marker="5.1.2"),
                ForgeDataMod(name="snad", marker="1.18.2-1.22.04.15a"),
                ForgeDataMod(name="jei", marker="9.7.0.209"),
                ForgeDataMod(name="ae2", marker="11.1.4"),
                ForgeDataMod(name="mekanism", marker="10.2.5"),
                ForgeDataMod(name="bdlib", marker="1.19.3.7"),
                ForgeDataMod(name="create", marker="0.5.0.d"),
                ForgeDataMod(name="waystones", marker="10.1.0"),
                ForgeDataMod(name="clumps", marker="8.0.0+10"),
                ForgeDataMod(name="shutupexperimentalsettings", marker="1.0.5"),
                ForgeDataMod(name="comforts", marker="1.18.2-5.0.0.4"),
                ForgeDataMod(name="naturescompass", marker="1.18.2-1.9.7-forge"),
                ForgeDataMod(name="storagenetwork", marker="1.18.2-1.6.1"),
                ForgeDataMod(name="framedcompactdrawers", marker="1.18-4.1.0"),
                ForgeDataMod(name="decorative_blocks", marker="2.1.0"),
                ForgeDataMod(name="botanypots", marker="8.0.12"),
                ForgeDataMod(name="ftbbackups2", marker="1.0.17"),
                ForgeDataMod(name="cofh_core", marker="1.6.4.21"),
                ForgeDataMod(name="mcjtylib", marker="1.18-6.0.15"),
                ForgeDataMod(name="ispawner", marker="1.0"),
                ForgeDataMod(name="everycomp", marker="1.18.2-1.5.7"),
                ForgeDataMod(name="jeitweaker", marker="3.0.0.8"),
                ForgeDataMod(name="terralith", marker="0.0NONE"),
                ForgeDataMod(name="mininggadgets", marker="1.11.0"),
                ForgeDataMod(name="crafttweaker", marker="9.1.197"),
                ForgeDataMod(name="akashictome", marker="1.5-20"),
                ForgeDataMod(name="forge", marker="ANY"),
                ForgeDataMod(name="colossalchests", marker="1.8.3"),
                ForgeDataMod(name="selene", marker="1.18.2-1.17.9"),
                ForgeDataMod(name="drippyloadingscreen", marker="1.6.4"),
                ForgeDataMod(
                    name="craftingtweaks",
                    marker="<not required for client>",
                ),
                ForgeDataMod(name="minecraft", marker="1.18.2"),
                ForgeDataMod(name="terrablender", marker="1.18.2-1.1.0.102"),
                ForgeDataMod(
                    name="sophisticatedbackpacksvh",
                    marker="1.18.2-1.0.4.12",
                ),
                ForgeDataMod(name="mousetweaks", marker="ANY"),
                ForgeDataMod(name="titanium", marker="3.5.6"),
                ForgeDataMod(name="jade", marker="<not required for client>"),
                ForgeDataMod(name="createtweaker", marker="2.0.0.17"),
                ForgeDataMod(name="easy_villagers", marker="1.18.2-1.0.10"),
                ForgeDataMod(name="pipez", marker="1.18.2-1.1.5"),
                ForgeDataMod(name="iceberg", marker="ANY"),
                ForgeDataMod(name="flywheel", marker="<not required for client>"),
                ForgeDataMod(name="mantle", marker="1.9.27"),
                ForgeDataMod(name="ecologics", marker="1.7.3"),
                ForgeDataMod(name="quark", marker="3.2-358"),
                ForgeDataMod(name="xaerominimap", marker="22.11.1"),
                ForgeDataMod(name="pigpen", marker="8.0.1"),
                ForgeDataMod(name="fastbench", marker="6.0.2"),
                ForgeDataMod(name="polymorph", marker="1.18.2-0.44"),
                ForgeDataMod(name="autoreglib", marker="1.7-53"),
                ForgeDataMod(name="storagedrawers", marker="10.2.1"),
                ForgeDataMod(name="fluxnetworks", marker="7.0.7.8"),
                ForgeDataMod(name="neoncraft2", marker="2.2"),
                ForgeDataMod(name="enercell", marker="0.0NONE"),
                ForgeDataMod(name="appleskin", marker="2.4.0+mc1.18"),
                ForgeDataMod(
                    name="ferritecore",
                    marker="<not required for client>",
                ),
                ForgeDataMod(name="modularrouters", marker="9.1.1-93"),
                ForgeDataMod(name="refinedstorageaddons", marker="0.8.2"),
                ForgeDataMod(name="openloader", marker="12.0.1"),
                ForgeDataMod(name="the_vault", marker="1.18.2-2.0.10.869"),
            ],
        ),
        ("truncated", False),
    ]

    @pytest.fixture(scope="class")
    def build(self) -> ForgeData:
        value = ForgeData.build(
            RawForgeData(
                {
                    "channels": [],
                    "d": (
                        bytes.fromhex(
                            "e0ba8b0000c484e4a0b0e18e9be19997e2baaee1b399e392bae7a5a6e6908ae4a2b8c5b1e380a3e2b1a1e1a39ee39eb6e"
                            "78db0e5bb86e19789e0a0b3e18ba3e49aa6e0b18be38686e685a8e5b39ce38695e6abbbe19896e2b78de0b181e1a097e3"
                            "80ae02d098e2aeabe19987e2b7ade0b181e1a097e38caee1b880e59684e4af83e19b86e4b0ade1b99ce398b1e68dafe69"
                            "b96e490b5e0a5b1e68e83e29985e0b08be1a097e384aed1a8e4b1a0ceabe29997e2b6aee1b298e392bae6b9aae6a1ace0"
                            "b388e78dbbe199a6e0b3ade1a99bcab1e2b8b1e5b1a2e38398e4ae98e39ba7e6aface1af98e38cb7e69da9cba6c384e4a"
                            "090e49890e0b2ade5b39ee39eb6e78da2e6888ce492b8e78781e48da2e2b6a1e1a998e2beb7e6a1a3e5b382e196b9e0ad"
                            "a3cc90e48080e1a184e386b9e6a5a8e4aba8e5868de7ae9be19c85e2b68ce1b499e38abae38485e6899ce4a2b8e48081e"
                            "198b0e2b3ace5b299e3aab4e0b1ade5b1a2e68384e185b1e18b93e29786e0ae8c18e6b48ae6bb86e2979de28db3e79bb6"
                            "e2b9aee0b281e1a097e38caee28884e3b78ce48e83e39a96e2ba8ee5a39ae3a8b0e691a5e5bb86e19789e0a28be18ba3e"
                            "49c86e4b28be1a096e394aee6999ce3a388e3a689e78e93e0b1a0e1a19ae39cb7e6b1a5e6888200e280b8e59a87e2b98c"
                            "e1a19bd0b6e2b8b1e5b1ace3a38ce48691e380a3e4b981e5b499e39eb7e78dace48b84e1978de0a193e18ba3e29c86e0b"
                            "38be1a097e3a4aee69096e58699e7adbbe39b86e2b18ee5b398c6b2e2b8b1c9a0e48080e78d88e49a96e2b4aee5ac98e3"
                            "9cb4e695b6e6a39ce4a6bde2af8be68da0e49885e0b88bdc81e789a9e5b39ee1969de2adb3e19ca6e6ba8ce5b29bcab9e"
                            "2b8b2e5b1a0e3a384e18d88e69bb7e2b3ade5ae99e3a4b2e791a1e6939ed78dc688e580a0e2bc81e1a598e39eb9e6bdb7"
                            "e5a3a4e39691cc8be181a7e49786e0b58ce1a297e6b484e58b82e0b6b9e78688e18c8240e5a385e39eb7e6a5abe4bb9ce"
                            "3b699e18e93e79b86e6b1ade5a89ae382b2e78da4e6888ce3a388e78681e78ca2e2b780e5b499e39ebbe6adb2e68886e4"
                            "82b8e0a081e382b0e4b7ace1b49be39eb9e6b1ace5b392e0a69de480a0e59ba7e4ba8ce0b182e1a297e2b4b8e5b1a8e3a"
                            "380e286a9e69e80e0b2ade4839de19c98c4b0e0b884e38780e1ac8be29996e2b7ace0b681e1a897e384aee6808ed6b1e2"
                            "ac9be798a6e282ade0ae8ce19c98c4b0e0b884e2968ce0aea3e59986e4b68ce0b181e1a297e2b8b1e1a1a6d6b4e78d8be"
                            "397b6e2b48ce1ae98e38ab7c5ac62e19080e7ae80e19db6e4b48ce0b382e1a097e384aee4919ae58695cc8be28290e6b7"
                            "ace5ab9be390b9e6b1a5e0bb8ce4b384e185b1e58ba3d886c980e39eb6e791afe4ab84e39685e38e9be68c90d8a5e4ae8"
                            "ec498e78c96e6839ee296a1e28e9be39a97e0b0ace1a59de384b2e68da1e68396e0b685e1ad9be184a7e29786e0b88ce1"
                            "a497e38cade6899ce3a3a0e2a699e78ba3e49aa6e4878ce390b1e6b9a1e4ab9cd6b1c688e58080e6ba80e5a99de3a6b3e"
                            "38493e6899ce582b8c5a9e49897e0b1aee4b19ae1a295e384aee5b1b0e0a388e181b8e19d96e0b68de1a999e38eb7e685"
                            "a7e4bb88e58695e48e9be68cb1e698a5e0ae8ce19a98e795a2e5a392e3a691e5a6a9e39b92e498ace0b18be19c9ce7b4b"
                            "2e5a888e29685e0adb3cd80e28080e5a482e3a4b0e795abe58ba8e4b6b1e0a0b3e68c83e49885e0b58bd080e68dade4a3"
                            "aee3b6bde1ae93e18197d786e0ae8c1be79c87e4a382e38691e1acabe18397e29786e0b88ce1a497e380ade7819ce492b"
                            "8e18789e584a0e6b2ade1a19ae392b7e6b5b3e4ab8ee196b9e0ae93e79d86e6b98de4869ce1a098e388aee6a99ce39188"
                            "e5acabe69896e6b4ade5ad9ce38ab3e695aee48ba4e3b791e1ae93e181a7e49886e0b28be1aa9700e49088e38685e3ada"
                            "be68cb0e49985e4b08be1a095d483e48baee386a5e58c8be59ba0e6ba8ce1af9de396b9e6b9a9e0ab8ee3a384e78681e1"
                            "8c82e68080e5aa82e3a4b2e78da5e6ab9ee0b789e1acabc2a7e29786e0b48ce1a297e384aee689aee38084e68c98e49bb"
                            "6e6b48ee5a397e39cb7e6a5a6d38ee4b0b8e2ad83e19d86e0b88de5a99ae39cb0e6bda3e4aba4e4b09de0a5b1e68c83d8"
                            "a5e48c8ce382b6e6b9a9e49abed6a1e78db3e49996e282ade0ae8ce19c9830e18884e59690e6adabe59b96e2b6ade1b99"
                            "be1a285e384aee5a9b0e3a384e786a9e48ca2e2b281e5ad9de3b2b6e6a1a3e5b382e196b9e0ada3cc90e68180e5b383e3"
                            "a0bae6b1b0e5ab8ae3a695e0aea3e19ca6e6b2ade48d9ce19c98e3a0b1e6919ce492b4e2a5b1e18ba3e699a6e5ae81e3a"
                            "8b2e6bdb7e59ba4e49085e18081e283a0e4b2aee1a999e38ab7e78da4e5bba8d789e2acbbe181a6e29786e0b08ce1a497"
                            "e6b48ce58b82e7b6b9e48c9ae69896e2b78de5ac99e1a28000e59890e3a6bde18d9be49997e2b2aee0b181e1a697e38ca"
                            "ee1a080d694e4ae9bd7b7e6b4aee5ac99e39cb4e0b1b3e5b1a2e68384e185b1e18b93d786e0ae8cc498e68c86e6939ee4"
                            "b781e68cabe68c90d8a5e0ae8ee19a99e2b8b1e5b1a0e1b388e2aca0e199a6e0baace5b49be1a282e380aee6819cd0844"
                            "0e39897e2b5ace1a59be3aab7e0a880e49b9ae0a79de4ae93e79986e6b2acc59ce19c99e2b8b066e580ace18dbbd8b7e2"
                            "b6ade1b398e38abad9b2e781a2e492b8c5b1e38083e6b1a1e1ad9be3a4b8e78da5e58ba6e39795e0a2b3e48ba3e49786e"
                            "0ad8ce3aab1e6b1a9e5b388e2b3a4e1ada9e68c96d8a5e0ae8ec499e78084e5b392e2a69de78688cc92e296a7e0ae8ce1"
                            "9c9ce0b0b0e58ba0e1b6b9e1abbbe19a86e4b78ce1a59bcab6e4a590e0ba9cd385e68090e29a90e4b7aee5a69be3a4bae"
                            "685aee4ab86e1978de78698e68cb2d8a5e1a985e39eb9e699aee693aad6b9e2ac9be79cb6e2b78be5b499e39ebbe6adb2"
                            "e68886e482b809e582b0e6b1ade1b49de3a0b9e6bda4e6939ee1978de78688e68c82db85e48980e386b6e699b7e5b38ae"
                            "1968de2ae9be68c90e49885e0b58be0ac80e795b3e4aba0e39789e18c8be19d87e4b58de0b69be1a49ae6bda3e4aba4e2"
                            "96b1e38c93e68c90e49885e4b18b1ce78c8be5ab92e38781e68f8be79a96e0b48ce4959de19c98e3a0b1e6919ce492b4e"
                            "285b1e28ba3e496a6e5b598e398b4e2b9a4e689a6e1b088e7ac90e19d86e2b78ce1a19ae1a285e384aee5b1b0e39388e1"
                            "a6a1e48d83e2b6a0e1a998c2b730e19880e296a0e48cbbe19b86e0b3ade5b49ae3a4b2e48483e38a9ce19085ce98e2989"
                            "7e4b5aee48680e3aab1e6a5b2e69b9ee490b9e0a5b1e68e83e29985e0b58be1a097e39caee6899ce39090e4ac8be19ba6"
                            "d8a0e48280e3a084e791a1e58386e596bde4ada3e182b6e29786e0b88ce1a497e39cade5b1a2e18384e0ada8e69a96e28"
                            "0ad0ccc81e685a3e4ab9ad789e0a1a3e18ba3e49c86e4b28be1a296e380aee6a19ce1809ce38cabe59896e0b68ee4859d"
                            "e19c98e2b8b0c9a0e3a080e68c90e39bb6e6b5ace1a198e3a0b9e6b9a5e693a8e2a7a5e78688cc92d6a7e4ae8ce19c993"
                            "0e6a0a4e196a1e6ae93e49896e4afade5af99e39cbae685a4e58ba8e3a6bde0a183e68ba3e69786e0ae8ce1b099e18480"
                            "e583a8e4a695e0adabe79b86e0b2abe5b09ee39cb0e6a5b3e5b39ee490a1e385b1e38ba3e29786e0b38cd681e6a5ace5b"
                            "384e3a6bde6af8be59bb6e2b9aee0b281e1a297e380aee4988ed6a1e78db3e49996e281ade0ae8cc298e0a882e5a38ae5"
                            "a695e28c8be29bb7e0b4aee48c99e19c98e3a0b1e6919ce492b4e485b1e48ba3e28686e5a19be39cb4e68d9fe48b90e3a"
                            "6b9e68cabe1809606c780e3aab9e695aee58b98e1a68de0a688cba3e29786e4858ce3ae84e789afe4a398e18695e28d8b"
                            "e380b7e2baace4819ac298e6a488e6a39ce4a695e0adb3e19b86e298a0e48280e38681e6b5a6e6b896e482b8c5b1cb93e"
                            "2b98ee4b299ce9ce695aee6bba8e4a6bde0ad9bcc9040e1a183e386b9e6a5a8e4aba8e5868de18eabe69e97e49a80e0b9"
                            "8be1b097dcb4e4ab9ce5b791e18dbbe19ab7e298a000e3ae87e6a5a5e4a3a4e3a6a5e3acbbe49896e2b3ace1b499e1a48"
                            "3e388aee6899cc384e6a1a8e798b6e2b38ee1b29de392b7e795b4e4aba4e1978de78698e68c82e49885c980e3a4bae78d"
                            "a1e49b90e3a685e38e9be68c90e49885e4b18bc89ae685ade5b392e4908501e58290e6b1ade5ac9de38eb4e791a8e0aba"
                            "6e3a384e78681ccb2e68480e5b598e3aab1e689ade6938ae59095e0a5b1e28ba306e1b381e382b7e189a4e5b1a2e68384"
                            "e185b1e18b93e49786e0b28ce1a097e2b8b4e6a9a2e0a684e58098e19996e284ade4ae8ee19c9be2b8b0e681a4e1b3a4e"
                            "48c98e69896e2b78de5ac99e1a282e380aee6819ce0a084e0a098e29996e28386e0b18ce1a297e390aee5a888e29685e0"
                            "adb3e18c9040e5ad82e396b2e6b9a1e69b92e1a6b5c688e28ba3e29786e4888de38ab6e685abe58b9ce3978de0a0b3e68"
                            "c83e49985e0b58bc880e68885e5a388e0a6a5e0a183e18ba3e49ca6e0b38be1ae97e6b48ae5a3aae29791e68c93e39bb6"
                            "e2b5ace0b280c880e6a5ade49ba6e49085e18081e381a0e2b98ce1a199e38abae38087e6a99ce482b8e285b1e58186e2b"
                            "0ade5ae9ae1a280c880e6b892e69685e28e9be69bb7e6b2ade4869ce1a098e384aee6819ce3a09ce28cabe79db7e6b98d"
                            "e4839ae19c9830e0b080e3868ce6aeabe39c86c48ee0ae8ee19c98e2acb0e681a2e6a080e48e98e49d96e0baaee1a59ce"
                            "3a0bce789a5e5ab92e3a695e0aea3e39b86e0b2aee5b49de39cb4e78da7e6888ae482b8e2a5b123e6b1a1e1ad9be39eb3"
                            "e791b2e1b3a6e3a384e48689e28ba3e296a6e0ae8de19c98e2b8b0e0a1a8d6b4e78d8be18096e48086ce80e382b7e795b"
                            "4e4aba4e0b78de6adbbe19c86e6b9ace4929ce19c98e3a0b1e6919ce492b4e4a5b1e78ba3e496a6e1af99e38eb9e0b9a5"
                            "e48b9ce59791e2ae93e39cb6e2b7ace5b09be3a6b0cdb3e5b1a2d380e78090e49cb0e4b7aee5a19ce38ab3e695aee6bba"
                            "8e4a6bde68d9be68c90d8a5e0ae8ee19a99e2b8b1e5b1ace38384e0ada8e69a96e6afade5a898e39cb0e695aecb98c384"
                            "e28080e299a1e2b0aee1a59be386b2e6b5afe48ba0e5868de18ca3e79897e4b2aee1b39ce1a285e384aee5a9b0e3a390e"
                            "78689cc82c8a0e5a599e39eb1e685b2e58ba8e19799e18bbbe79b86e6b1ade5b39ae1a482e384aee6819ce2a080e7ac90"
                            "e19d86e2b78ce5b09ee3a8b7d9b3e5b1b0e3a380e18689e38083e0b381e1a29de382b1e6ada3e683aae4a78de0a0b1cba"
                            "3e29786e0b78cd281e6bda3e5838ce0b5bde18dbbd997e498a1e0b68be1a897e388aee0b9a2e1969ce2adb3e19ca6e2b6"
                            "8ce4b180c480e6b488e59386e69791e4ada3e398a6e498a1e0b18be19a9ce2b8b6e5b1a0e59384e6a181e298b6e2ba8de"
                            "5ac9ee384b4e38483e6819c04e4a180dcb6e6b0aee5ae9de3a4b2e38483e6819ce29080e38ca8e29997e6bcaee5af98e3"
                            "a0b6e3848ce6899ce3a3a0e6a691e68c92e49aa5e0b78bd480e695aae6a392e1979de5ac8be29996e683aee0ae8ce19c9"
                            "8e2b8b070e580a4e18cabe19ca7e2b68ce1b49aceb4e2b8b0e1b1a0e3a4bde188aae58390e4b4ade1a99be38eb7e685a7"
                            "e4bb88e58695e38e9be68c90e298a5e0ae8ce0a898e685ade5b392e3a5bde28cabe79db7e6b98de59f9ae390b1e6b9a1e"
                            "4ab9cd6b1c690e480a0e4b1a1e1a19ce3a8b3e79db4e48b8ae196ade3ae93e68e90e498a5e4b18be1ae9ce6b484e58b82"
                            "e196b9e78688e68c82d885e48280e38285e685abe583a6e0b6a5e7aea3e59b96e2838ce4ae8ce19a9ae380b2e5a888e29"
                            "685e0adb3cc90e28280e5a681e3a4b7e695a7c886e694b9e281a2e59a97e6b98ce5b397e3a4b7e6a5b4e4bb9ce4908dc5"
                            "b1e58083e0b9a0e5ac9ce3a8b4e38483e6899ce0a084e1a1b0e49bb6e6b7ade5b39ce398b0e6a1a3e69b8ae4b791e0a0a"
                            "bcba3e69787e48c8ce390b1e6b9a1e4ab9ce7b6b1e0adaae69a96e282ade0ae8ce19c98c4b0e0b084e1978ce2ada3e59b"
                            "a6e286ace4ae8ce1b098e388aee6899ae492b8e786b9e78e92e2b780e5b499e39ebbe6adb2e6888200e28298e19ca6e0b"
                            "88de1b99ce39eb6e691a1e5b392e4b69de18c9be59997e2b78ce0b181e1ac97e390aee1b086e4a68ce38c8be19d86e6b7"
                            "8de5b499e38abbe6ada1e0bba6e196b8e3aea3e29bb7e6b5aee0b180e1a097d080e5a892e3a6a5e1acabe19ca6e0b38ce"
                            "4869de19c98e3a0b1e6919ce590a8e18db3e79997e6b4ace5b49ce3a4b2e49884e1a29ad38de18180e79997e6b4ace5b4"
                            "9ce3a4b2e49884e1a29ad38de68080e59d80e4b98ce1a19ce398b1e6b9a5e4ab88e48789e78688cc92e49787e4ad8ce19"
                            "c98e2b8b1e5b1a0e48384c691e38680e0b7aee5a89ce3a6b4e6a5b4e48b86e19791e18ca3e39896e0b5ace5a19ce396b1"
                            "e799b3e1bb90e3a384e48689e28ba3e296a6e0ae8ce19c98e2b8b4e691a2e2b080e7ada8e39d96e0b2aee5b79de382b2e"
                            "78dabc886e694b9e48092e19d80e2ba8de5ae98e3aab4d5ade5b1a6e3a394e3a6b1e59ba0e6ba8ce1af9de396b9e38483"
                            "e6819cd084e580a0e49896e0b2ace48d80e3a4b1e685a5e4aba8e5b791e0acabe59ab6e0b98ce0b282e1a097e380aee68"
                            "99ce0a39ce2a1b0e39896e6bcaee5b697e398b4e685ace4ab8ee4b789e0a1abe18ba3e49c86e4b28be1a296e380aee689"
                            "9ce1b380e2aca0e199a6e0baace5b49be1a282e380aee6819ce0a08428da97e4b2aee48c9ee19c98e3a0b1e6919ce492b"
                            "4e0a5b1e58ba3c3a6e1a599e382b3e6b1b5e0aba8e3a384e78681e18c82e68080e5a981e38ab1e695a2e4bba4d08de4a9"
                            "b215e0b381e5b99be390bbe695a5d398e39098e78c8be49d86e4b2ade0b181e1b297e388aee0b9aee196b8e3aea3e29bb"
                            "7e2b5aee0b18000e69489e5bb86e3b6b1e4acbbe398b6e282aee4ae8ce19c9bc8b3e6888ad795e5ae93e381b6e49786e4"
                            "ad8ce1aa99d0b8e48b9ae3a6a5e0a08be28083c680e5a19ee3a4b2e6b5afe5b392e396a5cc8be281b7e49986e4b18be19"
                            "c98d0b1e48b9ae3a6a5e0a09bcba326c680e392b8e781a7e5b38ae68095c5b1e18ba3e28186e5a682e3a6b0e689b4e5b3"
                            "8ae2868de380abcba3e49786e4878ce390b1e6b9a1e4ab9ce196b1e786a0e68da2e29885e48280e3a084e6b1afe5abb2e"
                            "4a6bde48e83e182b6e29786e0b88ce1a497e380ade6a19ce18390e0ada8e69a96e280ad0cd480e795a1e5bba8e19789e6"
                            "8cbbe29a96e2838ce4ae8ce19a9be38cb5e1b084e5878ce18dbbe79897e0b2ace5b299e3aeb0e789a5e0b3a6e48384e18"
                            "5b1e18ba3e28686e5a19be39cb4e68d9fe48b90e3a6b9e68cabe1809606cc80e398b3e7a1b5e4ab9ce5b791e18dbbe39a"
                            "b7e683aee0ae8de19c98e2b8b770e3a0a8e7acabe39ba6e2b98ce1a698e1a4bae38883e6919ce28088e78ca8e29996e2b"
                            "1aee1ac99ceb6e2b8b0e1b1a0e3a4bde3a8aae59ba0e6ba8ce1af9de396b9e38085e6819ce482b8e18081e18290e0b88c"
                            "e5ac9ce3a6b2e6a5abe1a39ce3a388e786a1e38c82e6b6a5e0b198e1a297d0b8e78ba6e0b6b9e0a08be18093e485a0e1a"
                            "599e3a4b9e791a9e49b8ae4a6bde18cabe583a0e0b7ade1b599e382b6e789b2e6ab9ee19791e1ae93e18287e29787e4ae"
                            "8ce19a98e38cb9e5a898e29685e7adb3d8b5e4b0ade5ae9be398b2e3888100e4a190e38cabe69a96e0b2ade1b399e39eb"
                            "ae685b2e4ab8ee18685e7aca3e39ba6c2aee0ae8ce19c9c32e5b894e19781e68db3e19bb6e2b28ce1b299e1a283e2b8b2"
                            "e5b1a0e0a384e28188e59a87e4aface5a19de398bae185b4e5b1a2e68384e185b1e28b93d786e4ae8ce1a098e3a0aee78"
                            "9ace3a09ce28cabe79db7e6b98dc69ae19c98e398b2e6819cd080e6a2a8e49bb6e4b2ace5ae9ce392bae698bae6ab98e3"
                            "a7a1e28cabe79db7e6b98de5b39ae1ae81e39cb000"
                        )
                    ).decode("utf-8"),
                    "fmlNetworkVersion": 3,
                    "mods": [],
                    "truncated": True,
                }
            )
        )
        assert value is not None
        return value



from __future__ import annotations

import dns.resolver
import sys
import json
import argparse
import socket
import dataclasses
from typing import TYPE_CHECKING
from typing_extensions import TypeAlias

from mcstatus import JavaServer, BedrockServer
from mcstatus.responses import JavaStatusResponse
from mcstatus.motd import Motd

if TYPE_CHECKING:
    SupportedServers: TypeAlias = "JavaServer | BedrockServer"

PING_PACKET_FAIL_WARNING = (
    "warning: contacting {address} failed with a 'ping' packet but succeeded with a 'status' packet,\n"
    "         this is likely a bug in the server-side implementation.\n"
    '         (note: ping packet failed due to "{ping_exc}")\n'
    "         for more details, see: https://mcstatus.readthedocs.io/en/stable/pages/faq/\n"
)

QUERY_FAIL_WARNING = (
    "The server did not respond to the query protocol."
    "\nPlease ensure that the server has enable-query turned on,"
    " and that the necessary port (same as server-port unless query-port is set) is open in any firewall(s)."
    "\nSee https://wiki.vg/Query for further information."
)


def _motd(motd: Motd) -> str:
    """Formats MOTD for human-readable output, with leading line break
    if multiline."""
    s = motd.to_ansi()
    return f"\n{s}" if "\n" in s else f" {s}"


def _kind(serv: SupportedServers) -> str:
    if isinstance(serv, JavaServer):
        return "Java"
    elif isinstance(serv, BedrockServer):
        return "Bedrock"
    else:
        raise ValueError(f"unsupported server for kind: {serv}")


def _ping_with_fallback(server: SupportedServers) -> float:
    # bedrock doesn't have ping method
    if isinstance(server, BedrockServer):
        return server.status().latency

    # try faster ping packet first, falling back to status with a warning.
    ping_exc = None
    try:
        return server.ping(tries=1)
    except Exception as e:
        ping_exc = e

    latency = server.status().latency

    address = f"{server.address.host}:{server.address.port}"
    print(
        PING_PACKET_FAIL_WARNING.format(address=address, ping_exc=ping_exc),
        file=sys.stderr,
    )

    return latency


def ping_cmd(server: SupportedServers) -> int:
    print(_ping_with_fallback(server))
    return 0


def status_cmd(server: SupportedServers) -> int:
    response = server.status()

    java_res = response if isinstance(response, JavaStatusResponse) else None

    if not java_res:
        player_sample = ""
    elif java_res.players.sample is not None:
        player_sample = str([f"{player.name} ({player.id})" for player in java_res.players.sample])
    else:
        player_sample = "No players online"

    if player_sample:
        player_sample = " " + player_sample

    print(f"version: {_kind(server)} {response.version.name} (protocol {response.version.protocol})")
    print(f"motd:{_motd(response.motd)}")
    print(f"players: {response.players.online}/{response.players.max}{player_sample}")
    print(f"ping: {response.latency:.2f} ms")
    return 0


def json_cmd(server: SupportedServers) -> int:
    data = {"online": False, "kind": _kind(server)}

    status_res = query_res = exn = None
    try:
        status_res = server.status(tries=1)
    except Exception as e:
        exn = exn or e

    try:
        if isinstance(server, JavaServer):
            query_res = server.query(tries=1)
    except Exception as e:
        exn = exn or e

    # construct 'data' dict outside try/except to ensure data processing errors
    # are noticed.
    data["online"] = bool(status_res or query_res)
    if not data["online"]:
        assert exn, "server offline but no exception?"
        data["error"] = str(exn)

    if status_res is not None:
        data["status"] = dataclasses.asdict(status_res)

        # ensure we are overwriting the motd and not making a new dict field
        assert "motd" in data["status"], "motd field missing. has it been renamed?"
        data["status"]["motd"] = status_res.motd.simplify().to_minecraft()

    if query_res is not None:
        # TODO: QueryResponse is not (yet?) a dataclass
        data["query"] = qdata = {}

        qdata["ip"] = query_res.raw["hostip"]
        qdata["port"] = query_res.raw["hostport"]
        qdata["map"] = query_res.map_name
        qdata["plugins"] = query_res.software.plugins
        qdata["raw"] = query_res.raw

    json.dump(data, sys.stdout)
    return 0


def query_cmd(server: SupportedServers) -> int:
    if not isinstance(server, JavaServer):
        print("The 'query' protocol is only supported by Java servers.", file=sys.stderr)
        return 1

    try:
        response = server.query()
    except socket.timeout:
        print(QUERY_FAIL_WARNING, file=sys.stderr)
        return 1

    print(f"host: {response.raw['hostip']}:{response.raw['hostport']}")
    print(f"software: {_kind(server)} {response.software.version} {response.software.brand}")
    print(f"motd:{_motd(response.motd)}")
    print(f"plugins: {response.software.plugins}")
    print(f"players: {response.players.online}/{response.players.max} {response.players.list}")
    return 0


def main(argv: list[str] = sys.argv[1:]) -> int:
    parser = argparse.ArgumentParser(
        "mcstatus",
        description="""
        mcstatus provides an easy way to query 1.7 or newer Minecraft servers for any
        information they can expose. It provides three modes of access: query, status,
        ping and json.
        """,
    )

    parser.add_argument("address", help="The address of the server.")
    parser.add_argument("--bedrock", help="Specifies that 'address' is a Bedrock server (default: Java).", action="store_true")

    subparsers = parser.add_subparsers(title="commands", description="Command to run, defaults to 'status'.")
    parser.set_defaults(func=status_cmd)

    subparsers.add_parser("ping", help="Ping server for latency.").set_defaults(func=ping_cmd)
    subparsers.add_parser("status", help="Prints server status.").set_defaults(func=status_cmd)
    subparsers.add_parser(
        "query", help="Prints detailed server information. Must be enabled in servers' server.properties file."
    ).set_defaults(func=query_cmd)
    subparsers.add_parser(
        "json",
        help="Prints server status and query in json.",
    ).set_defaults(func=json_cmd)

    args = parser.parse_args(argv)
    lookup = JavaServer.lookup if not args.bedrock else BedrockServer.lookup

    try:
        server = lookup(args.address)
        return args.func(server)
    except (socket.timeout, socket.gaierror, dns.resolver.NoNameservers, ConnectionError, TimeoutError) as e:
        # catch and hide traceback for expected user-facing errors
        print(f"Error: {e!r}", file=sys.stderr)
        return 1


if __name__ == "__main__":
    sys.exit(main())



import asyncio
import typing
from asyncio.exceptions import TimeoutError
from unittest.mock import patch

import pytest

from mcstatus.address import Address
from mcstatus.protocol.connection import TCPAsyncSocketConnection


class FakeAsyncStream(asyncio.StreamReader):
    async def read(self, *args, **kwargs) -> typing.NoReturn:
        await asyncio.sleep(2)
        raise NotImplementedError("tests are designed to timeout before reaching this line")


async def fake_asyncio_asyncio_open_connection(hostname: str, port: int):
    return FakeAsyncStream(), None


class TestAsyncSocketConnection:
    @pytest.mark.asyncio
    async def test_tcp_socket_read(self):
        with patch("asyncio.open_connection", fake_asyncio_asyncio_open_connection):
            async with TCPAsyncSocketConnection(Address("dummy_address", 1234), timeout=0.01) as tcp_async_socket:
                with pytest.raises(TimeoutError):
                    await tcp_async_socket.read(10)



from __future__ import annotations

import pytest

from mcstatus.motd import Motd
from mcstatus.motd.components import Formatting, MinecraftColor, TranslationTag, WebColor
from mcstatus.responses import RawJavaResponseMotdWhenDict


class TestMotdParse:
    def test_correct_result(self, source):
        assert Motd.parse(source) == Motd(
            [
                "top", Formatting.RESET,
                "1", Formatting.RESET,
                WebColor.from_hex(hex="#b3eeff"), "2", Formatting.RESET,
                MinecraftColor.BLACK, Formatting.OBFUSCATED, "3", Formatting.RESET,
                MinecraftColor.DARK_BLUE, Formatting.BOLD, Formatting.STRIKETHROUGH, "4", Formatting.RESET,
                MinecraftColor.DARK_GREEN, Formatting.ITALIC, "5", Formatting.RESET,
                MinecraftColor.DARK_AQUA, Formatting.UNDERLINED, "6", Formatting.RESET,
                MinecraftColor.DARK_AQUA, "7", Formatting.RESET,
                MinecraftColor.DARK_RED, "8", Formatting.RESET,
                MinecraftColor.DARK_PURPLE, "9", Formatting.RESET,
                MinecraftColor.GOLD, "10", Formatting.RESET,
                MinecraftColor.GRAY, "11", Formatting.RESET,
                MinecraftColor.DARK_GRAY, "12", Formatting.RESET,
                MinecraftColor.BLUE, "13", Formatting.RESET,
                MinecraftColor.GREEN, "14", Formatting.RESET,
                MinecraftColor.AQUA, "15", Formatting.RESET,
                MinecraftColor.RED, "16", Formatting.RESET,
                MinecraftColor.LIGHT_PURPLE, "17", Formatting.RESET,
                MinecraftColor.YELLOW, "18", Formatting.RESET,
                MinecraftColor.WHITE, "19", Formatting.RESET,
                MinecraftColor.MINECOIN_GOLD, "20", Formatting.RESET,
                Formatting.RESET, "21", Formatting.RESET,
                TranslationTag("some.random.string"), Formatting.RESET,
            ],
            raw=source,
        )  # fmt: skip

    @pytest.mark.parametrize("bedrock", (True, False))
    def test_bedrock_parameter_nothing_changes(self, bedrock: bool):
        assert Motd.parse([{"color": "minecoin_gold", "text": " "}], bedrock=bedrock).parsed == [
            Formatting.RESET,
            MinecraftColor.MINECOIN_GOLD,
            " ",
            Formatting.RESET,
        ]

    @pytest.mark.parametrize("bedrock,expected", ((True, MinecraftColor.MINECOIN_GOLD), (False, "&g")))
    def test_parse_as_str_ignore_minecoin_gold_on_java(self, bedrock: bool, expected):
        assert Motd.parse("&g", bedrock=bedrock).parsed == [expected]

    def test_parse_incorrect_color_passes(self):
        """See `https://github.com/py-mine/mcstatus/pull/335#discussion_r985084188`_."""
        assert Motd.parse("&j").parsed == ["&j"]

    def test_parse_uppercase_passes(self):
        assert Motd.parse("&A").parsed == ["", MinecraftColor.GREEN, ""]

    @pytest.mark.parametrize(
        "input,expected", [("", [""]), ([], [Formatting.RESET]), ({"extra": [], "text": ""}, ["", Formatting.RESET])]
    )
    def test_empty_input_also_empty_raw(self, input, expected):
        assert Motd.parse(input).parsed == expected

    def test_top_level_formatting_applies_to_all_in_extra(self) -> None:
        """As described `here <https://wiki.vg/Chat#Inheritance>`_."""
        assert Motd.parse({"text": "top", "bold": True, "extra": [{"color": "red", "text": "not top"}]}).parsed == [
            Formatting.BOLD,
            "top",
            Formatting.RESET,
            Formatting.BOLD,
            MinecraftColor.RED,
            "not top",
            Formatting.RESET,
        ]

    def test_top_level_formatting_can_be_overwrote(self) -> None:
        """As described `here <https://wiki.vg/Chat#Inheritance>`_."""
        assert Motd.parse(
            {"text": "bold", "bold": True, "extra": [{"color": "red", "bold": False, "text": "not bold"}]}
        ).parsed == [
            Formatting.BOLD,
            "bold",
            Formatting.RESET,
            MinecraftColor.RED,
            "not bold",
            Formatting.RESET,
        ]

    def test_top_level_formatting_applies_to_string_inside_extra(self) -> None:
        """Although, it is probably a bug in some modded cores, Minecraft supports it, and we should as well.

        See `#711 <https://github.com/py-mine/mcstatus/issues/711>`_.
        """
        assert Motd.parse({"text": "top", "bold": True, "extra": ["not top"]}).parsed == [
            Formatting.BOLD,
            "top",
            Formatting.RESET,
            Formatting.BOLD,
            "not top",
        ]

    def test_formatting_key_set_to_false_here_without_it_being_set_to_true_before(self) -> None:
        """Some servers set the formatting keys to false here, even without it ever being set to true before.

        See `https://github.com/py-mine/mcstatus/pull/335#discussion_r985086953`_.
        """
        assert Motd.parse({"color": "red", "bold": False, "text": "not bold"}).parsed == [
            MinecraftColor.RED,
            "not bold",
            Formatting.RESET,
        ]

    def test_translate_string(self):
        assert Motd.parse(RawJavaResponseMotdWhenDict(translate="the key")).parsed == [
            TranslationTag("the key"),
            Formatting.RESET,
        ]

    def test_short_text_is_not_considered_as_color(self):
        """See `https://github.com/py-mine/mcstatus/pull/335#discussion_r984535349`_."""
        assert Motd.parse("a").parsed == ["a"]

    def test_text_field_contains_formatting(self):
        """See `https://github.com/py-mine/mcstatus/pull/335#issuecomment-1264191303`_."""
        assert Motd.parse({"text": "&aHello!"}).parsed == ["", MinecraftColor.GREEN, "Hello!", Formatting.RESET]

    def test_invalid_raw_input(self):
        with pytest.raises(TypeError):
            Motd.parse(object())  # type: ignore

    def test_invalid_color(self):
        with pytest.raises(ValueError):
            Motd._parse_color("a")

    def test_multiple_times_nested_extras(self):
        """See `https://discord.com/channels/936788458939224094/938591600160956446/1062860329597534258`_."""
        motd = Motd.parse(
            {
                "extra": [
                    {
                        "extra": [
                            {"extra": [{"text": "1"}]},
                            {"extra": [{"text": "2"}]},
                            {"extra": [{"text": "3"}]},
                        ]
                    },
                    {
                        "extra": [
                            {"extra": [{"text": "4"}]},
                            {"extra": [{"text": "5"}]},
                            {"extra": [{"text": "6"}]},
                        ]
                    },
                    {
                        "extra": [
                            {"extra": [{"text": "7"}]},
                            {"extra": [{"text": "8"}]},
                            {"extra": [{"text": "9"}]},
                        ]
                    },
                ]
            }
        )
        assert motd.parsed == [
            Formatting.RESET, Formatting.RESET, Formatting.RESET,
            "1",
            Formatting.RESET, Formatting.RESET,
            "2",
            Formatting.RESET, Formatting.RESET,
            "3",
            Formatting.RESET, Formatting.RESET, Formatting.RESET,
            "4",
            Formatting.RESET, Formatting.RESET,
            "5",
            Formatting.RESET, Formatting.RESET,
            "6",
            Formatting.RESET, Formatting.RESET, Formatting.RESET,
            "7",
            Formatting.RESET, Formatting.RESET,
            "8",
            Formatting.RESET, Formatting.RESET,
            "9",
            Formatting.RESET,
        ]  # fmt: skip

    def test_raw_attribute(self, source):
        motd = Motd.parse(source)
        assert motd.raw == source



from __future__ import annotations

import ipaddress
import sys
from pathlib import Path
from typing import cast
from unittest.mock import MagicMock, Mock, patch

import dns.resolver
import pytest
from dns.rdatatype import RdataType

from mcstatus.address import Address, async_minecraft_srv_address_lookup, minecraft_srv_address_lookup


class TestSRVLookup:
    @pytest.mark.parametrize("exception", [dns.resolver.NXDOMAIN, dns.resolver.NoAnswer])
    def test_address_no_srv(self, exception):
        with patch("dns.resolver.resolve") as resolve:
            resolve.side_effect = [exception]
            address = minecraft_srv_address_lookup("example.org", default_port=25565, lifetime=3)
            resolve.assert_called_once_with("_minecraft._tcp.example.org", RdataType.SRV, lifetime=3, search=True)

        assert address.host == "example.org"
        assert address.port == 25565

    def test_address_with_srv(self):
        with patch("dns.resolver.resolve") as resolve:
            answer = Mock()
            answer.target = "different.example.org."
            answer.port = 12345
            resolve.return_value = [answer]

            address = minecraft_srv_address_lookup("example.org", lifetime=3)
            resolve.assert_called_once_with("_minecraft._tcp.example.org", RdataType.SRV, lifetime=3, search=True)
        assert address.host == "different.example.org"
        assert address.port == 12345

    @pytest.mark.asyncio
    @pytest.mark.parametrize("exception", [dns.resolver.NXDOMAIN, dns.resolver.NoAnswer])
    async def test_async_address_no_srv(self, exception):
        with patch("dns.asyncresolver.resolve") as resolve:
            resolve.side_effect = [exception]
            address = await async_minecraft_srv_address_lookup("example.org", default_port=25565, lifetime=3)
            resolve.assert_called_once_with("_minecraft._tcp.example.org", RdataType.SRV, lifetime=3, search=True)

        assert address.host == "example.org"
        assert address.port == 25565

    @pytest.mark.asyncio
    async def test_async_address_with_srv(self):
        with patch("dns.asyncresolver.resolve") as resolve:
            answer = Mock()
            answer.target = "different.example.org."
            answer.port = 12345
            resolve.return_value = [answer]

            address = await async_minecraft_srv_address_lookup("example.org", lifetime=3)
            resolve.assert_called_once_with("_minecraft._tcp.example.org", RdataType.SRV, lifetime=3, search=True)
        assert address.host == "different.example.org"
        assert address.port == 12345


class TestAddressValidity:
    @pytest.mark.parametrize(
        "address,port",
        [
            ("example.org", 25565),
            ("192.168.0.100", 54321),
            ("2345:0425:2CA1:0000:0000:0567:5673:23b5", 100),
            ("2345:0425:2CA1::0567:5673:23b5", 12345),
        ],
    )
    def test_address_validation_valid(self, address, port):
        Address._ensure_validity(address, port)

    @pytest.mark.parametrize(
        "address,port,exception",
        [
            # Out of range port
            ("example.org", 100_000, ValueError),
            ("example.org", -1, ValueError),
            # Invalid types
            ("example.org", "25565", TypeError),
            (25565, "example.org", TypeError),
            (("example.org", 25565), None, TypeError),
            (0, 0, TypeError),
            ("", "", TypeError),
        ],
    )
    def test_address_validation_invalid(self, address, port, exception):
        with pytest.raises(exception):
            Address._ensure_validity(address, port)


class TestAddressConstructing:
    def test_init_constructor(self):
        addr = Address("example.org", 25565)
        assert addr.host == "example.org"
        assert addr.port == 25565

    def test_tuple_behavior(self):
        addr = Address("example.org", 25565)
        assert isinstance(addr, tuple)
        assert len(addr) == 2
        assert addr[0] == "example.org"
        assert addr[1] == 25565

    def test_from_tuple_constructor(self):
        addr = Address.from_tuple(("example.org", 12345))
        assert addr.host == "example.org"
        assert addr.port == 12345

    def test_from_path_constructor(self):
        addr = Address.from_path(Path("example.org:25565"))
        assert addr.host == "example.org"
        assert addr.port == 25565

    def test_address_with_port_no_default(self):
        addr = Address.parse_address("example.org:25565")
        assert addr.host == "example.org"
        assert addr.port == 25565

    def test_address_with_port_default(self):
        addr = Address.parse_address("example.org:25565", default_port=12345)
        assert addr.host == "example.org"
        assert addr.port == 25565

    def test_address_without_port_default(self):
        addr = Address.parse_address("example.org", default_port=12345)
        assert addr.host == "example.org"
        assert addr.port == 12345

    def test_address_without_port(self):
        with pytest.raises(ValueError):
            Address.parse_address("example.org")

    def test_address_with_invalid_port(self):
        with pytest.raises(ValueError):
            Address.parse_address("example.org:port")

    def test_address_with_multiple_ports(self):
        with pytest.raises(ValueError):
            Address.parse_address("example.org:12345:25565")


class TestAddressIPResolving:
    def setup_method(self):
        self.host_addr = Address("example.org", 25565)
        self.ipv4_addr = Address("1.1.1.1", 25565)
        self.ipv6_addr = Address("::1", 25565)

    def test_ip_resolver_with_hostname(self):
        with patch("dns.resolver.resolve") as resolve:
            answer = MagicMock()
            cast(MagicMock, answer.__str__).return_value = "48.225.1.104."
            resolve.return_value = [answer]

            resolved_ip = self.host_addr.resolve_ip(lifetime=3)

            resolve.assert_called_once_with(self.host_addr.host, RdataType.A, lifetime=3, search=True)
            assert isinstance(resolved_ip, ipaddress.IPv4Address)
            assert str(resolved_ip) == "48.225.1.104"

    @pytest.mark.asyncio
    async def test_async_ip_resolver_with_hostname(self):
        with patch("dns.asyncresolver.resolve") as resolve:
            answer = MagicMock()
            cast(MagicMock, answer.__str__).return_value = "48.225.1.104."
            resolve.return_value = [answer]

            resolved_ip = await self.host_addr.async_resolve_ip(lifetime=3)

            resolve.assert_called_once_with(self.host_addr.host, RdataType.A, lifetime=3, search=True)
            assert isinstance(resolved_ip, ipaddress.IPv4Address)
            assert str(resolved_ip) == "48.225.1.104"

    def test_ip_resolver_with_ipv4(self):
        with patch("dns.resolver.resolve") as resolve:
            resolved_ip = self.ipv4_addr.resolve_ip(lifetime=3)

            resolve.assert_not_called()  # Make sure we didn't needlessly try to resolve
            assert isinstance(resolved_ip, ipaddress.IPv4Address)
            assert str(resolved_ip) == self.ipv4_addr.host

    @pytest.mark.asyncio
    async def test_async_ip_resolver_with_ipv4(self):
        with patch("dns.asyncresolver.resolve") as resolve:
            resolved_ip = await self.ipv4_addr.async_resolve_ip(lifetime=3)

            resolve.assert_not_called()  # Make sure we didn't needlessly try to resolve
            assert isinstance(resolved_ip, ipaddress.IPv4Address)
            assert str(resolved_ip) == self.ipv4_addr.host

    def test_ip_resolver_with_ipv6(self):
        with patch("dns.resolver.resolve") as resolve:
            resolved_ip = self.ipv6_addr.resolve_ip(lifetime=3)

            resolve.assert_not_called()  # Make sure we didn't needlessly try to resolve
            assert isinstance(resolved_ip, ipaddress.IPv6Address)
            assert str(resolved_ip) == self.ipv6_addr.host

    @pytest.mark.asyncio
    async def test_async_ip_resolver_with_ipv6(self):
        with patch("dns.asyncresolver.resolve") as resolve:
            resolved_ip = await self.ipv6_addr.async_resolve_ip(lifetime=3)

            resolve.assert_not_called()  # Make sure we didn't needlessly try to resolve
            assert isinstance(resolved_ip, ipaddress.IPv6Address)
            assert str(resolved_ip) == self.ipv6_addr.host

    def test_resolve_localhost(self):
        addr = Address("localhost", 25565)

        context_manager = pytest.warns(RuntimeWarning) if sys.platform == "darwin" else MagicMock()
        with context_manager:
            assert addr.resolve_ip() == ipaddress.ip_address("127.0.0.1")

    @pytest.mark.asyncio
    async def test_async_resolve_localhost(self):
        addr = Address("localhost", 25565)

        context_manager = pytest.warns(RuntimeWarning) if sys.platform == "darwin" else MagicMock()
        with context_manager:
            assert await addr.async_resolve_ip() == ipaddress.ip_address("127.0.0.1")



import pytest

from mcstatus.motd import Motd
from mcstatus.responses import JavaStatusPlayer, JavaStatusPlayers, JavaStatusResponse, JavaStatusVersion
from tests.responses import BaseResponseTest


@BaseResponseTest.construct
class TestJavaStatusResponse(BaseResponseTest):
    RAW = {
        "players": {"max": 20, "online": 0},
        "version": {"name": "1.8-pre1", "protocol": 44},
        "description": "A Minecraft Server",
        "enforcesSecureChat": True,
        "favicon": "data:image/png;base64,foo",
    }

    EXPECTED_VALUES = [
        ("players", JavaStatusPlayers(0, 20, None)),
        ("version", JavaStatusVersion("1.8-pre1", 44)),
        ("motd", Motd.parse("A Minecraft Server", bedrock=False)),
        ("latency", 0),
        ("enforces_secure_chat", True),
        ("icon", "data:image/png;base64,foo"),
        ("raw", RAW),
        ("forge_data", None),
    ]
    OPTIONAL_FIELDS = (
        [("favicon", "icon"), ("enforcesSecureChat", "enforces_secure_chat")],
        {
            "players": {"max": 20, "online": 0},
            "version": {"name": "1.8-pre1", "protocol": 44},
            "description": "A Minecraft Server",
            "enforcesSecureChat": True,
            "favicon": "data:image/png;base64,foo",
        },
    )

    @pytest.fixture(scope="class")
    def build(self) -> JavaStatusResponse:
        return JavaStatusResponse.build(self.RAW)  # type: ignore # dict[str, Unknown] cannot be assigned to TypedDict

    def test_as_dict(self, build: JavaStatusResponse):
        assert build.as_dict() == {
            "enforces_secure_chat": True,
            "forge_data": None,
            "icon": "data:image/png;base64,foo",
            "latency": 0,
            "motd": "A Minecraft Server",
            "players": {"max": 20, "online": 0, "sample": None},
            "raw": {
                "description": "A Minecraft Server",
                "enforcesSecureChat": True,
                "favicon": "data:image/png;base64,foo",
                "players": {"max": 20, "online": 0},
                "version": {"name": "1.8-pre1", "protocol": 44},
            },
            "version": {"name": "1.8-pre1", "protocol": 44},
        }


@BaseResponseTest.construct
class TestJavaStatusPlayers(BaseResponseTest):
    EXPECTED_VALUES = [
        ("max", 20),
        ("online", 0),
        (
            "sample",
            [
                JavaStatusPlayer("foo", "0b3717c4-f45d-47c8-b8e2-3d9ff6f93a89"),
                JavaStatusPlayer("bar", "61699b2e-d327-4a01-9f1e-0ea8c3f06bc6"),
                JavaStatusPlayer("baz", "40e8d003-8872-412d-b09a-4431a5afcbd4"),
            ],
        ),
    ]
    OPTIONAL_FIELDS = (
        [("sample", "sample")],
        {
            "max": 20,
            "online": 0,
            "sample": [
                {"name": "foo", "id": "0b3717c4-f45d-47c8-b8e2-3d9ff6f93a89"},
                {"name": "bar", "id": "61699b2e-d327-4a01-9f1e-0ea8c3f06bc6"},
                {"name": "baz", "id": "40e8d003-8872-412d-b09a-4431a5afcbd4"},
            ],
        },
    )

    @pytest.fixture(scope="class")
    def build(self) -> JavaStatusPlayers:
        return JavaStatusPlayers.build(
            {
                "max": 20,
                "online": 0,
                "sample": [
                    {"name": "foo", "id": "0b3717c4-f45d-47c8-b8e2-3d9ff6f93a89"},
                    {"name": "bar", "id": "61699b2e-d327-4a01-9f1e-0ea8c3f06bc6"},
                    {"name": "baz", "id": "40e8d003-8872-412d-b09a-4431a5afcbd4"},
                ],
            }
        )

    def test_empty_sample_turns_into_empty_list(self) -> None:
        assert JavaStatusPlayers.build({"max": 20, "online": 0, "sample": []}).sample == []


@BaseResponseTest.construct
class TestJavaStatusPlayer(BaseResponseTest):
    EXPECTED_VALUES = [("name", "foo"), ("id", "0b3717c4-f45d-47c8-b8e2-3d9ff6f93a89")]

    @pytest.fixture(scope="class")
    def build(self) -> JavaStatusPlayer:
        return JavaStatusPlayer.build({"name": "foo", "id": "0b3717c4-f45d-47c8-b8e2-3d9ff6f93a89"})

    def test_id_field_the_same_as_uuid(self) -> None:
        unique = object()
        build = JavaStatusPlayer.build({"name": "foo", "id": unique})  # type: ignore[assignment]
        assert build.id is build.uuid
        assert build.uuid is unique


@BaseResponseTest.construct
class TestJavaStatusVersion(BaseResponseTest):
    EXPECTED_VALUES = [("name", "1.8-pre1"), ("protocol", 44)]

    @pytest.fixture(scope="class")
    def build(self) -> JavaStatusVersion:
        return JavaStatusVersion.build({"name": "1.8-pre1", "protocol": 44})



from pytest import fixture, mark

from mcstatus.motd import Motd
from mcstatus.responses import BedrockStatusPlayers, BedrockStatusResponse, BedrockStatusVersion
from tests.responses import BaseResponseTest


@fixture(scope="module")
def build():
    return BedrockStatusResponse.build(
        [
            "MCPE",
            "§r§4G§r§6a§r§ey§r§2B§r§1o§r§9w§r§ds§r§4e§r§6r",
            "422",
            "1.18.100500",
            "1",
            "69",
            "3767071975391053022",
            "map name here",
            "Default",
            "1",
            "19132",
            "-1",
            "3",
        ],
        123.0,
    )


@BaseResponseTest.construct
class TestBedrockStatusResponse(BaseResponseTest):
    EXPECTED_VALUES = [
        ("motd", Motd.parse("§r§4G§r§6a§r§ey§r§2B§r§1o§r§9w§r§ds§r§4e§r§6r", bedrock=True)),
        ("latency", 123.0),
        ("map_name", "map name here"),
        ("gamemode", "Default"),
    ]
    EXPECTED_TYPES = [
        ("players", BedrockStatusPlayers),
        ("version", BedrockStatusVersion),
    ]

    @fixture(scope="class")
    def build(self, build):  # pyright: ignore[reportIncompatibleMethodOverride]
        return build

    @mark.parametrize("field,pop_index", [("map_name", 7), ("gamemode", 7), ("gamemode", 8)])
    def test_optional_parameters_is_none(self, field, pop_index):
        parameters = [
            "MCPE",
            "§r§4G§r§6a§r§ey§r§2B§r§1o§r§9w§r§ds§r§4e§r§6r",
            "422",
            "1.18.100500",
            "1",
            "69",
            "3767071975391053022",
            "map name here",
            "Default",
        ]
        parameters.pop(pop_index)
        # remove all variables after `pop_index`
        if len(parameters) - 1 == pop_index:
            parameters.pop(pop_index)

        build = BedrockStatusResponse.build(parameters, 123.0)
        assert getattr(build, field) is None

    def test_as_dict(self, build: BedrockStatusResponse):
        assert build.as_dict() == {
            "gamemode": "Default",
            "latency": 123.0,
            "map_name": "map name here",
            "motd": "§4G§6a§ey§2B§1o§9w§ds§4e§6r",
            "players": {"max": 69, "online": 1},
            "version": {"brand": "MCPE", "name": "1.18.100500", "protocol": 422},
        }


@BaseResponseTest.construct
class TestBedrockStatusPlayers(BaseResponseTest):
    EXPECTED_VALUES = [("online", 1), ("max", 69)]

    @fixture(scope="class")
    def build(self, build):  # pyright: ignore[reportIncompatibleMethodOverride]
        return build.players


@BaseResponseTest.construct
class TestBedrockStatusVersion(BaseResponseTest):
    EXPECTED_VALUES = [("name", "1.18.100500"), ("protocol", 422), ("brand", "MCPE")]

    @fixture(scope="class")
    def build(self, build):  # pyright: ignore[reportIncompatibleMethodOverride]
        return build.version



import pytest


@pytest.fixture(scope="session")
def source() -> dict:
    """Returns ultimate dict with almost all possible aspects, which we should support.

    If feature can handle all from this dict, it's fully tested.
    Parser should have more tests, on additional features.
    """
    return {
        "extra": [
            {"text": "1"},
            {"color": "#b3eeff", "text": "2"},
            {"obfuscated": True, "color": "black", "text": "3"},
            {"bold": True, "strikethrough": True, "color": "dark_blue", "text": "4"},
            {"italic": True, "color": "dark_green", "text": "5"},
            {"underlined": True, "color": "dark_aqua", "text": "6"},
            {"color": "dark_aqua", "text": "7"},
            {"color": "dark_red", "text": "8"},
            {"color": "dark_purple", "text": "9"},
            {"color": "gold", "text": "10"},
            {"color": "gray", "text": "11"},
            {"color": "dark_gray", "text": "12"},
            {"color": "blue", "text": "13"},
            {"color": "green", "text": "14"},
            {"color": "aqua", "text": "15"},
            {"color": "red", "text": "16"},
            {"color": "light_purple", "text": "17"},
            {"color": "yellow", "text": "18"},
            {"color": "white", "text": "19"},
            {"color": "minecoin_gold", "text": "20"},
            {"color": "reset", "text": "21"},
            {"translate": "some.random.string"},
        ],
        "text": "top",
    }
